"""
Discretization of DFE to SFS transformation.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import logging
from functools import cached_property, wraps
from typing import Literal, Sequence, Tuple

import mpmath as mp
import numpy as np
from scipy.integrate import quad

from .optimization import parallelize as parallelize_func
from .parametrization import Parametrization

# configure logger
logger = logging.getLogger('fastdfe').getChild('Discretization')


def get_midpoints_and_spacing(bins: np.ndarray) -> (np.ndarray, np.ndarray):
    """
    Obtain midpoints and spacing for the given bins.

    :param bins: Array of bins
    :return: Midpoints, spacing
    """
    # obtain midpoints
    s = (bins[1:] + bins[:-1]) / 2
    interval_sizes = bins[1:] - bins[:-1]

    return s, interval_sizes


def get_bins(intervals_del: tuple, intervals_ben: tuple) -> np.ndarray:
    """
    Get bins using log-spaced distances for positive and negative
    values of S given by ``intervals_ben`` and ``intervals_del``, respectively.

    :param intervals_del: ``(min, max, n)`` for negative values of S where we use log-spaced distances
    :param intervals_ben: ``(min, max, n)`` for positive values of S where we use log-spaced distances
    :return: Bins
    """
    bins_del = -np.logspace(np.log10(np.abs(intervals_del[0])),
                            np.log10(np.abs(intervals_del[1])), intervals_del[2] + 1)

    bins_ben = np.logspace(np.log10(intervals_ben[0]), np.log10(intervals_ben[1]), intervals_ben[2] + 1)

    return np.concatenate([bins_del, bins_ben])


def H(x, S):
    """
    Allele frequency sojourn time. Note that this function is not used in practice.

    :param S: Selection coefficient
    :param x: Allele frequency
    :return: Allele frequency sojourn time
    """
    return ((1 - np.exp(-S * (1 - x), dtype=np.float128)) /
            (x * (1 - x) * (1 - np.exp(-S, dtype=np.float128)))).astype(float)


def H_regularized(x, S: float | int | np.ndarray):
    """
    As H(x, S) but replacing with the limits close to the limit points.
    Note that this function is not used in practice.

    :param x: Allele frequency
    :param S: Selection coefficient
    :return: Allele frequency sojourn time
    """
    # make it accept scalars
    if isinstance(S, (float, int)):
        return H_regularized(x, np.array([S]))[0]

    # else S is an array
    y = np.zeros_like(S, dtype=np.float128)

    # S close to zero
    close_to_zero = np.abs(S) < 1e-8

    # limit (1 - exp(-S * (1 - x))) / (x * (1 - x) * (1 - exp(-S))) as S -> 0
    # cf. https://www.wolframalpha.com/input?i=limit+%281+-+exp%28-S+*+%281+-+x%29%29%29+%2F+%
    # 28x+*+%281+-+x%29+*+%281+-+exp%28-S%29%29%29+as+S+-%3E+0
    y[close_to_zero] = 1 / x

    # evaluate function as usual
    y[~close_to_zero] = H(x, S[~close_to_zero])

    return y


def H_fixed(S: float | int | np.ndarray) -> float | int | np.ndarray:
    """
    The sojourn time as x -> 1. Note that this function is not used in practice.

    :param S: Selection coefficient
    :return: Sojourn time as x -> 1
    """
    return S / (1 - np.exp(-S))


def H_fixed_regularized(S: float | int | np.ndarray) -> float | int | np.ndarray:
    """
    As :func:`H_fixed` but replacing with the limits close to the limit points.
    Note that this function is not used in practice.

    :param S: Selection coefficient
    :return: Sojourn time as x -> 1
    """
    # make it accept scalars
    if isinstance(S, (float, int)):
        return H_fixed(np.array([S]))[0]

    # else S is an array
    y = np.zeros_like(S)

    # S close to zero
    close_to_zero = np.abs(S) < 1e-8

    # limit as S -> 0
    y[close_to_zero] = 1

    # avoid overflow for very low S
    very_negative = S < -500

    # for very low S we are close to 0
    y[very_negative] = 0

    # remaining values
    remaining = ~(close_to_zero | very_negative)

    # evaluate as usual
    y[remaining] = H_fixed(S[remaining])

    return y


def hyp1f1(a: float | np.ndarray, b: float | np.ndarray, z: float | np.ndarray) -> float | np.ndarray:
    """
    Vectorized version of mpmath's hyp1f1 function.

    :param a: First parameter
    :param b: Second parameter
    :param z: Third parameter
    :return: Result
    """
    return np.vectorize(mp.hyp1f1)(a, b, z)


def exp(x: float | np.ndarray) -> float | np.ndarray:
    """
    Vectorized version of mpmath's exp function.

    :param x: Exponent
    :return: Result
    """
    return np.vectorize(mp.exp)(x)


def to_float(func):
    """
    Decorator to convert the return value of a function to float.

    :param func: Function to be decorated
    :return: Wrapper function
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> float | np.ndarray:
        """
        Wrapper function.

        :param args: Positional arguments
        :param kwargs: Keyword arguments
        :return: Converted result
        """
        result = func(*args, **kwargs)

        if isinstance(result, np.ndarray):
            return result.astype(float)
        else:
            return float(result)

    return wrapper


class Discretization:
    """
    The integral mapping the DFE to allele counts can be discretized
    and computed once beforehand so that transforming the DFE to
    the expected SFS counts can be carried out by matrix multiplication.
    To evaluate the integral over the specified intervals, midpoint
    integration or Scipy's quad method can be used.
    The integral can also alternatively be evaluated numerically
    in each optimizing iteration.
    The loss function supports Poisson likelihoods and the L2 norm.
    """

    def __init__(
            self,
            n: int,
            intervals_del: Tuple[float, float, int] = (-1.0e+8, -1.0e-5, 1000),
            intervals_ben: Tuple[float, float, int] = (1.0e-5, 1.0e4, 1000),
            integration_mode: Literal['midpoint', 'quad'] = 'midpoint',
            linearized: bool = True,
            parallelize: bool = True,
    ):
        """
        Create Discretization instance.

        :return: Number of individuals
        :param intervals_del: ``(start, stop, n_interval)`` for deleterious population-scaled selection coefficients.
        :param intervals_ben: ``(start, stop, n_interval)`` for beneficial population-scaled selection coefficients.
        :param integration_mode : 'midpoint' or 'quad' for midpoint integration or Scipy's quad method.
        :return: Whether to linearize the integral
        :param parallelize: Whether to parallelize the computation of the discretization
        """
        self.n = n

        # make sure lower bounds are lower than upper bounds
        if not intervals_del[0] < intervals_del[1] or not intervals_ben[0] < intervals_ben[1]:
            raise Exception('Lower intervals bounds must be lower than upper bounds.')

        # make sure |S| is not too large
        if intervals_del[0] < -1e10 or intervals_ben[1] > 1e10:
            raise Exception('Bounds for S should within the interval [-1e10, 1e10] to avoid '
                            'unexpected numerical behavior.')

        # bounds for discretizing DFE
        self.intervals_del = intervals_del
        self.intervals_ben = intervals_ben

        self.integration_mode = integration_mode
        self.linearized = linearized
        self.parallelize = parallelize

        # iteration counter
        self.n_it = 0

        # define bins, midpoints and step size
        # these intervals are used when linearizing the integral
        # and when plotting the DFE
        self.bins = get_bins(intervals_del, intervals_ben)
        self.s, self.interval_sizes = get_midpoints_and_spacing(self.bins)

        # the number of intervals
        self.n_intervals = self.s.shape[0]

    def get_dfe_to_sfs_midpoint(self) -> np.ndarray:
        """
        Precompute linearized integral using midpoint integration.
        Consider parallelizing this.

        :return: Matrix of size (n_intervals, n)
        """
        logger.info('Precomputing linear DFE-SFS transformation using midpoint integration.')

        I = np.ones((self.n - 1, self.n_intervals + 1))

        # we use midpoint integration
        K = np.arange(1, self.n)[:, None] * I
        S = self.bins[None, :] * I

        def compute_slice(i: int) -> np.ndarray:
            """
            Compute allele counts of a given multiplicity.

            :param i: Multiplicity
            :return: Discretized counts
            """
            return self.get_allele_count_regularized(S[i], K[i])

        # retrieve allele counts
        P = parallelize_func(
            func=compute_slice,
            data=np.arange(self.n - 1),
            parallelize=self.parallelize,
            desc=f"{self.__class__.__name__}>Precomputing",
            dtype=float
        )

        # take midpoint and multiply by interval size
        return (P[:, :-1] + P[:, 1:]) / 2 * self.interval_sizes

    def get_dfe_to_sfs_quad(self) -> np.ndarray:
        """
        Precompute linearized integral using Scipy's quad method.

        :return: Matrix of size (n_intervals, n)
        """
        logger.info('Precomputing linear DFE-SFS transformation using scipy.integrate.quad.')

        # initialize matrix
        P = np.zeros((self.bins.shape[0] - 1, self.n - 1))

        # iterate over bins
        for i in range(self.bins.shape[0] - 1):

            logger.debug(f"Processing interval {(self.bins[i], self.bins[i + 1])}.")

            # iterate over allele count classes
            for j in range(self.n - 1):
                P[i, j] = quad(lambda s: (self.get_allele_count_regularized(
                    s, j + 1)), self.bins[i], self.bins[i + 1])[0]

        # For GammaExpParametrization, there is a discontinuity at 0
        # which can cause problems integrating it using quadrature.
        # We remove this discontinuity here by taking the average value
        # of the adjacent bins.
        # Find bin that contains 0.
        i_zero = np.where((self.bins[:-1] < 0) & (self.bins[1:] >= 0))[0][0]

        # take average of adjacent bins
        P[i_zero] = (P[i_zero - 1] + P[i_zero + 1]) / 2

        return P.T

    @cached_property
    def dfe_to_sfs(self) -> np.ndarray:
        """
        Linearized DFE to SFS transformation.

        :return: Matrix of size (n_intervals, n)
        """
        # precompute linearized transformation.
        if self.integration_mode == 'midpoint':
            return self.get_dfe_to_sfs_midpoint()

        if self.integration_mode == 'quad':
            return self.get_dfe_to_sfs_quad()

        raise NotImplementedError(f'Integration mode {self.integration_mode} not supported.')

    @to_float
    def get_allele_count_large_negative_S(self, S: float | np.ndarray, k: float | np.ndarray) -> np.ndarray | float:
        """
        Limit of get_allele_count for large negative S.

        :param S: Population-scaled selection coefficient
        :param k: Allele count
        :return: Number of counts in frequency class P(k)
        """
        return self.n / (k * (self.n - k)) * hyp1f1(k, self.n, S)

    def get_allele_count_large_positive_S(self, S: float | np.ndarray, k: float | np.ndarray) -> np.ndarray | float:
        """
        Limit of get_allele_count for large positive S.

        :param S: Population-scaled selection coefficient
        :param k: Allele count
        :return: Number of counts in frequency class P(k)
        """
        return self.n / (k * (self.n - k))

    @to_float
    def get_allele_count(self, S: float | np.ndarray, k: float | np.ndarray) -> np.ndarray | float:
        """
        The number of counts in frequency class P(k) with
        population-scaled selection coefficient S.
        Binomial sampling is included here.
        See appendix of https://pubmed.ncbi.nlm.nih.gov/31975166

        :param S: Population-scaled selection coefficient
        :param k: Allele count
        :return: Number of counts in frequency class P(k)
        """
        return (self.n / (k * (self.n - k))) * ((1 - exp(-S) * hyp1f1(k, self.n, S)) / (1 - exp(-S)))

    def get_allele_count_regularized(self, S: float | np.ndarray, k: float | np.ndarray) -> float | np.ndarray:
        """
        As :meth:`get_allele_count` but using the respective limits
        for ``S`` close to zero and ``S`` very negative.
        Note that ``S`` and ``k`` need to have the same shape.

        :param S: Population-scaled selection coefficient
        :param k: Allele count
        :return: Number of counts in frequency class ``P(k)`
        """
        # make it accept scalars for S
        if isinstance(S, (float, int)):
            return self.get_allele_count_regularized(np.array([S]), np.array([k]))[0]

        # else S is an array
        y = np.zeros_like(S, dtype=float)

        # S close to zero
        close_to_zero = np.abs(S) < 1e-8

        # simply take limit value as S -> 0
        # see https://www.wolframalpha.com/input?i=limit+%28n%2F%28k%28n-k%29%29%29+*+
        # 1%2F%281-exp%28-S%29%29+*+%281-exp%28-S%29*1F1%28k%2Cn%2CS%29%29+as+S+-%3E+0
        if close_to_zero.any():
            y[close_to_zero] = np.ones_like(S[close_to_zero]) / k[close_to_zero]

        # large negative S
        very_negative = S < -1e4

        # Evaluate using limit.
        # We check if there are any negative values here
        # as numpy's vectorize does not work with empty arrays.
        if very_negative.any():
            y[very_negative] = self.get_allele_count_large_negative_S(S[very_negative], k[very_negative])

        # remaining values of S
        remaining = np.array(~(close_to_zero | very_negative))

        # evaluate function as usual for remaining values
        if remaining.any():
            y[remaining] = self.get_allele_count(S[remaining], k[remaining])

        return y

    def model_selection_sfs(self, model: Parametrization, params: dict) -> np.ndarray:
        """
        Infer SFS from the given DFE PDF.
        Note that demography is not included here.

        :param params: Parameters of the DFE
        :param model: DFE parametrization
        :return: SFS counts
        """
        if self.linearized:

            # get discretized DFE
            dfe = model._discretize(params, self.bins) / self.interval_sizes

            # get SFS from DFE using linearization
            # the interval sizes are already included here
            counts_modelled = self.dfe_to_sfs @ dfe
        else:
            dfe_pdf = model.get_pdf(**params)

            def integrate(k: int) -> float:
                """
                Integrate over DFE using scipy's quad.
                Here we use the PDF.
                """

                def integrate_dfe(s: float) -> float:
                    """
                    Integrate over DFE for fixed k.
                    Note that this produces too small results for low s,
                    but we don't make use of it in production code anyway.
                    """
                    return self.get_allele_count_regularized(s, k) * dfe_pdf(s)

                return quad(integrate_dfe, self.bins[0], self.bins[-1])[0]

            counts_modelled = np.array([integrate(k) for k in range(1, self.n)])

        return counts_modelled

    def get_alpha(self, model: Parametrization, params: dict) -> float:
        """
        Get alpha, the proportion of beneficial non-synonymous substitutions.

        :param model: DFE parametrization
        :param params: Parameters of the DFE
        :return: Estimated for alpha
        """
        y = model._discretize(params, self.bins) * H_fixed_regularized(self.s)

        return np.sum(y[self.s > 0]) / np.sum(y)

    def get_interval_density(
            self,
            inter: Sequence = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf])
    ) -> np.ndarray:
        """
        Get interval density.

        :param inter: Intervals
        :return: Interval density
        """
        return np.array([np.sum((inter[i - 1] < self.s) & (self.s < inter[i])) for i in range(1, len(inter))])



"""
DFE parametrizations.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import logging
import warnings
from abc import abstractmethod, ABC
from functools import wraps
from typing import Callable, List, Union, Dict, Tuple, Literal, Sequence

import numpy as np
from scipy.stats import gamma, expon

# get logger
logger = logging.getLogger('fastdfe')


def _from_string(model: Union['Parametrization', str]) -> 'Parametrization':
    """
    Return Parametrization from class name string.

    :param model: Class name string or parametrization object
    :return: Parametrization object
    """
    if isinstance(model, Parametrization):
        return model

    if isinstance(model, str):
        return globals()[model]()

    raise ValueError(f'Unknown parametrization: {model}')


def _to_string(model: Union['Parametrization', str]) -> str:
    """
    Return class name string from Parametrization.

    :param model: Class name string or Parametrization object
    :return: Class name string
    """
    if isinstance(model, Parametrization):
        return model.__class__.__name__

    return model


class Parametrization(ABC):
    """
    Base class for DFE parametrizations.

    Note that :func:`get_pdf` is not required to be implemented, provided that the
    linearized mode of fastDFE is used (which is highly recommended).
    """

    #: Default initial parameters
    x0: Dict[str, float] = {}

    #: Default parameter bounds
    bounds: Dict[str, Tuple[float, float]] = {}

    #: Scales over which to optimize the parameters, either 'log' or 'lin'
    scales: Dict[str, Literal['lin', 'log']] = {}

    #: The kind of submodels supported by holding some parameters fixed
    submodels: Dict[str, Dict[str, float]] = dict(
        full=dict(),
        dele=dict()
    )

    def __init__(self):
        """
        Initialize parametrization.
        """
        self._logger = logger.getChild(self.__class__.__name__)

        #: argument names
        self.param_names: List = list(self.x0.keys())

    @staticmethod
    def _accepts_scalars(func: Callable) -> Callable[[np.ndarray | float], np.ndarray | float]:
        """
        Make func accept scalar values.

        :return: Function that accepts scalars
        """

        @wraps(func)
        def wrapper(S: np.ndarray, *args, **kwargs) -> np.ndarray:
            """
            Wrapper function.

            :param S: Array of selection coefficients
            :param args: Positional arguments
            :param kwargs: Keyword arguments
            :return: Output of func
            """
            if isinstance(S, (int, float)):
                return func(np.array([S]), *args, **kwargs)[0]
            else:
                return func(S, *args, **kwargs)

        return wrapper

    @abstractmethod
    def get_pdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get probability distribution function of DFE.

        :return: Function that accepts an array of selection coefficients and returns their probability density
        """
        pass

    @abstractmethod
    def get_cdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get probability distribution function of DFE.

        :return: Function that accepts an array of selection coefficients and returns their cumulative probability
        """
        pass

    def _discretize(
            self,
            params,
            bins: np.ndarray,
            warn_mass: bool = True
    ) -> np.ndarray:
        """
        Discretize by using the CDF.

        :param params: Parameters of the parametrization
        :param bins: Bins to use for discretization
        :param warn_mass: Whether to warn if the total mass is not near 1.
        :return: Histogram
        """
        x = self.get_cdf(**params)(bins)

        # issue warning if values at bounds are outside [0, 1]
        if warn_mass and (not -1e-5 < x[0] < 1e-5 or not 1 - 1e-5 < x[-1] < 1 + 1e-5):
            self._logger.warning(f'CDF evaluates to {(x[0], x[-1])} at the lower and upper '
                                 f'bounds, which is a bit off from the expected (0, 1). '
                                 f'Used parameters: {params}.')

        # compute histogram
        hist = x[1:] - x[:-1]

        return hist

    def _normalize(self, params: dict) -> dict:
        """
        Normalize parameters.

        :param params: Parameters of the parametrization
        :return: Normalized parameters
        """
        # do nothing by default
        return params

    def plot(
            self,
            params: dict,
            intervals: Sequence = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            file: str = None,
            show: bool = True,
            title: str = 'discretized DFE',
            ax: 'plt.Axes' = None

    ) -> 'plt.Axes':
        """
        Plot the discretized DFE.

        :param params: Parameters of the parametrization
        :param intervals: Intervals to use for discretization.
        :param file: File to save the plot to.
        :param show: Whether to show the plot.
        :param title: Title of the plot.
        :param ax: Axes to use for the plot.
        :return: Axes
        """
        from .visualization import Visualization

        values = self._discretize(params, np.array(intervals), warn_mass=False)

        # check for nan values
        if np.isnan(values).any():
            self._logger.warning(f'NaN values in discretized DFE. Are the parameters valid?')

        return Visualization.plot_discretized(
            values=[values],
            file=file,
            show=show,
            intervals=np.array(intervals),
            title=title,
            ax=ax
        )


class GammaExpParametrization(Parametrization):
    r"""
    Parametrization for mixture of a gamma and exponential distribution. This corresponds to
    model C in polyDFE.

    We have the following probability density function:

    .. math::
        \phi(S; S_d, b, p_b, S_b) = (1 - p_b) f_\Gamma(-S; S_d, b) \cdot \mathbf{1}_{\{S < 0\}} +
        p_b f_e(S; S_b) \cdot \mathbf{1}_{\{S \geq 0\}}

    where:

    * :math:`S_d` is the mean of the DFE for :math:`S < 0`
    * :math:`b` is the shape of the gamma distribution
    * :math:`p_b` is the probability that :math:`S \geq 0`
    * :math:`S_b` is the mean of the DFE for :math:`S \geq 0`
    * :math:`f_\Gamma(x; m, b)` is the density of the gamma distribution with mean :math:`m` and shape :math:`b`
    * :math:`f_e(x; m)` is the density of the Exponential distribution with mean :math:`m`
    * :math:`\mathbf{1}_{\{A\}}` denotes the indicator function, which is 1 if :math:`A` is true, and 0 otherwise.

    The DFE has often been observed to be multi-modal for negative selection coefficients. A gamma distribution
    provides a good amount of flexibility to accommodate this. Note that all :math:`S` parameters are population-scaled
    selection coefficients, i.e. :math:`S = 4N_es`.

    """

    #: Default initial parameters
    x0: Dict[str, float] = dict(
        S_d=-1000,
        b=0.4,
        p_b=0.05,
        S_b=1
    )

    #: Default parameter bounds, using non-zero lower bounds for S_d and S_b due to log-scaled scales
    bounds: Dict[str, Tuple[float, float]] = dict(
        S_d=(-1e5, -1e-2),
        b=(0.01, 10),
        p_b=(0, 0.5),
        S_b=(1e-4, 100)
    )

    #: Scales over which to optimize the parameters
    scales: Dict[str, Literal['lin', 'log']] = dict(
        S_d='log',
        b='log',
        p_b='lin',
        S_b='log'
    )

    #: The kind of submodels supported by holding some parameters fixed
    submodels: Dict[str, Dict[str, float]] = dict(
        full=dict(),
        dele=dict(
            p_b=0,
            S_b=1
        )
    )

    @staticmethod
    def get_pdf(S_d: float, b: float, p_b: float, S_b: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get PDF.

        :param S_d: Mean selection coefficient for deleterious mutations
        :param b: Shape parameter for gamma distribution
        :param p_b: Probability of a beneficial mutation
        :param S_b: Mean selection coefficient for beneficial mutations
        :return: Function that accepts an array of selection coefficients and returns their probability density
        """

        @Parametrization._accepts_scalars
        def pdf(S: np.ndarray) -> np.ndarray:
            """
            The PDF.

            :param S: Selection coefficients
            :return: Probability density
            """
            x = np.zeros_like(S, dtype=float)

            # The gamma distribution may approach infinity as S -> 0
            # so in order to evaluate this function at s = 0,
            # we use--unlike polyDFE--the exponential mixture
            # distribution at this point.
            negative = S < 0

            # positive S
            x[negative] = (1 - p_b) * gamma.pdf(-S[negative], b, scale=-S_d / b)

            # Allow for S_b == 0 when p_b == 0 as well
            # which would produce an error otherwise.
            if S_b == 0 and p_b == 0:
                x[~negative] = 0
            else:
                # non-negative S
                x[~negative] = p_b * expon.pdf(S[~negative], scale=S_b)

            return x

        return pdf

    @staticmethod
    def get_cdf(S_d: float, b: float, p_b: float, S_b: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get CDF.

        :param S_d: Mean selection coefficient for deleterious mutations
        :param b: Shape parameter for gamma distribution
        :param p_b: Probability of a beneficial mutation
        :param S_b: Mean selection coefficient for beneficial mutations
        :return: Function that accepts an array of selection coefficients and returns their cumulative probability
        """

        @Parametrization._accepts_scalars
        def cdf(S: np.ndarray) -> np.ndarray:
            """
            The CDF.

            :param S: Selection coefficients
            :return: Cumulative probability
            """
            x = np.zeros_like(S, dtype=float)

            negative = S < 0

            # positive S
            x[negative] = (1 - p_b) - ((1 - p_b) * gamma.cdf(-S[negative], b, scale=-S_d / b))

            # Allow for S_b = 0 when p_b = 0 as well
            # which would produce an error otherwise.
            if S_b == 0 and p_b == 0:
                x[~negative] = 1
            else:
                # non-negative S
                x[~negative] = (1 - p_b) + p_b * expon.cdf(S[~negative], scale=S_b)

            return x

        return cdf


class DisplacedGammaParametrization(Parametrization):
    r"""
    Parametrization for a reflected displaced gamma distribution.

    We have the following probability density function:

    .. math::
        \phi(S; \hat{S}, b, S_{max}) = f_\Gamma(S_{max} - S; S_{max} - \hat{S}, b) \cdot \mathbf{1}_{\{S \leq S_{max}\}}

    where:

    * :math:`\hat{S}` is the mean of the DFE
    * :math:`b` is the shape of the gamma distribution
    * :math:`S_{max}` is the maximum value that :math:`S` can take
    * :math:`f_\Gamma(x; m, b)` is the density of the gamma distribution with mean :math:`m` and shape :math:`b`
    * :math:`\mathbf{1}_{\{A\}}` denotes the indicator function, which is 1 if :math:`A` is true, and 0 otherwise.

    This parametrization uses a single gamma distribution for both positive and negative selection coefficients.
    This is a less flexible parametrization, which may produce results similar to the other models while requiring
    fewer parameters. Note that all :math:`S` parameters are population-scaled selection coefficients, i.e.
    :math:`S = 4N_es`.

    .. warning::
        This model does not allow for a purely deleterious sub-parametrization, so
        :meth:`Inference.compare_nested_models` won't work as expected.

    """

    #: Default initial parameters
    x0: Dict[str, float] = dict(
        S_mean=-100,
        b=1,
        S_max=1
    )

    #: Default parameter bounds
    bounds: Dict[str, Tuple[float, float]] = dict(
        S_mean=(-100000, -0.01),
        b=(0.01, 10),
        S_max=(0.001, 100)
    )

    #: Scales over which to optimize the parameters
    scales: Dict[str, Literal['lin', 'log']] = dict(
        S_mean='log',
        b='lin',
        S_max='log'
    )

    #: The kind of submodels supported by holding some parameters fixed
    submodels: Dict[str, Dict[str, float]] = dict(
        full=dict(),
        dele=dict()
    )

    @staticmethod
    def get_pdf(S_mean: float, b: float, S_max: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get PDF.

        :return: Function that accepts an array of selection coefficients and returns their probability density
        """

        @Parametrization._accepts_scalars
        def pdf(S: np.ndarray) -> np.ndarray:
            """
            The PDF.

            :param S: Selection coefficients
            :return: Probability density
            """
            x = np.zeros_like(S, dtype=float)

            # calculate the probability density for S <= S_max
            is_lower = S <= S_max
            x[is_lower] = gamma.pdf(S_max - S[is_lower], b, scale=max(S_max - S_mean, 1e-16) / b)

            return x

        return pdf

    @staticmethod
    def get_cdf(S_mean: float, b: float, S_max: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get CDF.

        :return: Function that accepts an array of selection coefficients and returns their cumulative probability
        """

        @Parametrization._accepts_scalars
        def cdf(S: np.ndarray) -> np.ndarray:
            """
            The CDF.

            :param S: Selection coefficients
            :return: Cumulative probability
            """
            x = np.zeros_like(S, dtype=float)

            # calculate the cumulative probability for S <= S_max
            is_lower = S <= S_max
            x[is_lower] = 1 - gamma.cdf(S_max - S[is_lower], b, scale=max(S_max - S_mean, 1e-16) / b)

            # set the cumulative probability to 1 for S > S_max
            x[~is_lower] = 1

            return x

        return cdf


class GammaDiscreteParametrization(Parametrization):
    r"""
    Parametrization for a mixture of a gamma and discrete distribution. This corresponds to polyDFE's model B.

    We have the following probability density function:

    .. math::
        \phi(S; S_d, b, p_b, S_b) = (1 - p_b) f_{\Gamma}(S; S_d, b) \cdot \mathbf{1}_{\{S < 0\}} +
        p_b \cdot S_b \cdot \mathbf{1}_{\{0 \leq S \leq 1 / S_b\}}

    where:

    * :math:`S_d` is the mean of the DFE for :math:`S < 0`
    * :math:`b` is the shape of the gamma distribution
    * :math:`p_b` is the probability that :math:`S \geq 0`
    * :math:`S_b` is the shared selection coefficient of all positively selected mutations up to :math:`1/S_b`
    * :math:`f_\Gamma(x; m, b)` is the density of the gamma distribution with mean :math:`m` and shape :math:`b`

    This parametrization is similar to :class:`GammaExpParametrization`, but uses a constant mass for positive
    selection coefficients. The results should be rather similar in most cases. Note that all :math:`S` parameters
    are population-scaled selection coefficients, i.e. :math:`S = 4N_es`.
    """

    #: Default initial parameters
    x0: Dict[str, float] = dict(
        S_d=-1000,
        b=0.4,
        p_b=0.05,
        S_b=1
    )

    #: default parameter bounds
    bounds: Dict[str, Tuple[float, float]] = dict(
        S_d=(-1e5, -1e-2),
        b=(0.01, 10),
        p_b=(0, 0.5),
        S_b=(1e-4, 100)
    )

    #: scales over which to optimize the parameters
    scales: Dict[str, Literal['lin', 'log']] = dict(
        S_d='log',
        b='log',
        p_b='lin',
        S_b='log'
    )

    #: The kind of submodels supported by holding some parameters fixed
    submodels: Dict[str, Dict[str, float]] = dict(
        full=dict(),
        dele=dict(
            p_b=0,
            S_b=1
        )
    )

    @staticmethod
    def get_pdf(S_d: float, b: float, p_b: float, S_b: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get PDF.

        :param S_d: Mean of the DFE for S < 0
        :param b: Shape of the gamma distribution
        :param p_b: Probability that S > 0
        :param S_b: Shared selection coefficient of all positively selected mutations up to S_b
        :return: Function that accepts an array of selection coefficients and returns their probability density
        """

        @Parametrization._accepts_scalars
        def pdf(S: np.ndarray) -> np.ndarray:
            """
            The PDF.

            :param S: Selection coefficients
            :return: Probability density
            """
            x = np.zeros_like(S, dtype=float)

            negative = S < 0
            x[negative] = (1 - p_b) * gamma.pdf(-S[negative], b, scale=-S_d / b)

            equal_to_S_b = (0 <= S) & (S <= 1 / S_b)
            x[equal_to_S_b] = p_b * S_b

            # the density is 0 for all other cases, no need to explicitly set it

            return x

        return pdf

    @staticmethod
    def get_cdf(S_d: float, b: float, p_b: float, S_b: float, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get CDF.

        :param S_d: Mean of the DFE for S < 0
        :param b: Shape of the gamma distribution
        :param p_b: Probability that S > 0
        :param S_b: Shared selection coefficient of all positively selected mutations up to S_b
        :return: Function that accepts an array of selection coefficients and returns their cumulative probability
        """

        @Parametrization._accepts_scalars
        def cdf(S: np.ndarray) -> np.ndarray:
            """
            The CDF.

            :param S: Selection coefficients
            :return: Cumulative probability
            """
            x = np.zeros_like(S, dtype=float)

            negative = S < 0
            x[negative] = (1 - p_b) - ((1 - p_b) * gamma.cdf(-S[negative], b, scale=-S_d / b))

            within_S_b = (0 <= S) & (S <= 1 / S_b)
            x[within_S_b] = (1 - p_b) + p_b * S_b * S[within_S_b]

            greater_than_S_b = S > 1 / S_b
            x[greater_than_S_b] = 1

            return x

        return cdf


class DiscreteParametrization(Parametrization):
    r"""
    Parametrization for a discrete distribution. This corresponds to polyDFE's model D.
    By default we use 6 bins, but this can be changed by passing a different array to the constructor.
    The resulting parameter names are :math:`S_1, S_2, \dots, S_k`, where :math:`k` is the number of bins.

    That is, the probability density function is given by:

    :math:`\phi(S; S_1, S_2, \dots, S_k) = \sum_{i=1}^{k} S_i/c_i \cdot \mathbf{1}_{\{S \in B_i\}}`
    such that :math:`\sum_{i=1}^{k} S_i = 1,`

    where :math:`B_i` and :math:`c_i` are interval :math:`i` and the width of interval :math:`i`, respectively.

    This parametrization has the advantage of not imposing a shape on the DFE. For a reasonably fine parametrization,
    the number of parameters is larger than those of the other models, however. We generally also observe larger
    confidence intervals for this parametrization, and the optimization procedure may well be less efficient as
    we have to re-normalize the parameters to make sure they sum up to 1. Note that all :math:`S` parameters are
    population-scaled selection coefficients, i.e. :math:`S = 4N_es`.
    """

    def __init__(
            self,
            intervals: Sequence = np.array([-100000, -100, -10, -1, 0, 1, 1000])
    ):
        """
        Constructor.

        :param intervals: The intervals of the discrete distribution.
        """
        super().__init__()

        #: Intervals
        self.intervals: np.ndarray = np.concatenate(([-np.inf], intervals, [np.inf]))

        #: Interval sizes
        self.interval_sizes: np.ndarray = self.intervals[1:] - self.intervals[:-1]

        #: Number of intervals, including the two infinite ones
        self.k: int = self.intervals.shape[0] - 1

        #: All parameter names, including the fixed ones
        self.params: np.ndarray = np.array([f"S{i}" for i in range(self.k)])

        #: Parameter names that are not fixed
        self.param_names: List[str] = self.params[1:-1].tolist()

        #: Fixed parameters
        self.fixed_params: Dict[str, int] = {self.params[0]: 0, self.params[-1]: 0}

        #: Default initial parameters
        self.x0: Dict[str, float] = dict((p, 1 / (self.k - 2)) for p in self.param_names)

        #: Default parameter bounds
        self.bounds: Dict[str, Tuple[float, float]] = dict((p, (0, 1)) for p in self.param_names)

        #: Scales
        # noinspection all
        self.scales: Dict[str, Literal['lin', 'log']] = dict((p, 'lin') for p in self.param_names)

        #: Submodels
        self.submodels: Dict[str, Dict[str, float]] = dict(
            full=dict(),
            dele=dict((p, 0) for p in self.params[1:-1][self.intervals[1:-2] < 0])
        )

    def _normalize(self, params: dict) -> dict:
        """
        Add params for boundaries and normalize so that
        the parameters sum up to 1.

        :param params: Parameter values
        :return: Dict of normalized values plus remaining parameters
        """
        # make sure we only include parameter used for this parametrization
        filtered = dict((k, v) for k, v in params.items() if k in self.param_names)

        # sum up
        s = np.sum(list(filtered.values()))

        # normalize and include remaining parameters
        return params | dict((k, v / s) for k, v in filtered.items())

    def get_pdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get PDF.

        :param kwargs: Parameter values S1, S2, ..., Sk
        :return: Function that computes the PDF for given S values
        """

        # normalize and add boundary parameters
        values = self._normalize(kwargs) | self.fixed_params

        @Parametrization._accepts_scalars
        def pdf(S: np.ndarray) -> np.ndarray:
            """
            The PDF.

            :param S: Selection coefficients
            :return: Probability density
            """
            x = np.zeros_like(S, dtype=float)

            # iterate over parameters and assign values
            for i, p in enumerate(self.params):
                x[(self.intervals[i] < S) & (S <= self.intervals[i + 1])] = values[p] / self.interval_sizes[i]

            return x

        return pdf

    def get_cdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get CDF.

        :param kwargs: Parameter values S1, S2, ..., Sk
        :return: Function that computes the CDF for given S values
        """

        # normalize and add boundary parameters
        values = self._normalize(kwargs) | self.fixed_params

        # convert to numpy arrays
        intervals = np.array(self.intervals)
        interval_sizes = np.array(self.interval_sizes)

        # convert to ordered array values
        vals = np.array([values[self.params[i]] for i in range(self.k)])

        @Parametrization._accepts_scalars
        def cdf(S: np.ndarray) -> np.ndarray:
            """
            The CDF.

            :param S: Selection coefficients
            :return: Cumulative probability
            """
            # iterate over parameters and assign values
            cum = np.cumsum([values[self.params[i]] for i in range(self.k)])

            # obtain bin indices
            i = np.sum(self.intervals[:, None] <= S[None, :], axis=0) - 1

            # make sure we don't go out of bounds which can happen if S is np.inf
            i[i >= self.k] = self.k - 1

            # cumulative probability up to previous bin
            cum_prev = cum[np.maximum(i - 1, np.zeros_like(S, dtype=int))]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)

                # probability in current bin
                cum_within = np.abs(intervals[i] - S) / interval_sizes[i] * vals[i]

            cum_within[np.isnan(cum_within)] = 0

            # return cumulative probability
            return cum_prev + cum_within

        return cdf


class DiscreteFractionalParametrization(Parametrization):
    r"""
    Same model as :class:`DiscreteParametrization`, but re-parametrized by
    :math:`\hat{S}_1, \hat{S}_2, \dots, \hat{S}_{k-1}`, so that the mass in the ith interval :math:`S_i`
    is determined by the sum of masses to the left, i.e.

    :math:`S_i = \hat{S}_i \sum_{j<i} S_j, i = 1, \dots, k - 1`,

    :math:`S_k = 1 - \sum_{j=1}^{k-1} S_j`.

    This parametrization has the advantage of not imposing a shape on the DFE. For a reasonably fine parametrization,
    the number of parameters is larger than those of the other models, however. It is more easily optimized than
    :class:`DiscreteParametrization` as it has one parameter less but its parameters are more difficult to interpret.
    One disadvantage with discrete parametrizations is that there may be `gaps` in the estimated DFE. Note that all
    :math:`S` parameters are population-scaled selection coefficients, i.e. :math:`S = 4N_es`.
    """

    def __init__(
            self,
            intervals: Sequence = np.array([-100000, -100, -10, -1, 0, 1, 1000])
    ):
        """
        Constructor.

        :param intervals: The intervals of the discrete distribution.
        """
        super().__init__()

        #: Intervals
        self.intervals: np.ndarray = np.concatenate(([-np.inf], intervals, [np.inf]))

        #: Interval sizes
        self.interval_sizes: np.ndarray = self.intervals[1:] - self.intervals[:-1]

        #: Number of intervals, including the two infinite ones
        self.k: int = self.intervals.shape[0] - 1

        #: All parameter names, including fixed parameters
        self.params: np.ndarray = np.array([f"S{i}" for i in range(self.k)])

        #: Parameter names that are not fixed
        self.param_names: List[str] = self.params[1:-2].tolist()

        #: Fixed parameters
        self.fixed_params = {self.params[0]: 0, self.params[-2]: 1, self.params[-1]: 0}

        #: Default initial parameters
        self.x0: Dict[str, float] = self.to_fractional(dict((p, 1 / (self.k - 2)) for p in self.param_names))

        #: Default parameter bounds
        self.bounds: Dict[str, Tuple[float, float]] = dict((p, (0, 1)) for p in self.param_names)

        #: Scales
        # noinspection all
        self.scales: Dict[str, Literal['lin', 'log']] = dict((p, 'lin') for p in self.param_names)

        #: Submodels
        self.submodels: Dict[str, Dict[str, float]] = dict(
            full=dict(),
            dele=dict((p, 0) for p in self.params[1:-1][self.intervals[1:-2] < 0])
        )

    def to_nominal(self, params: Dict[str, float]) -> Dict[str, float]:
        """
        Convert representation of fraction of total mass to the left
        to representation of fractions which sum to 1.

        :param params: Parameter values S1, S2, ..., Sk
        :return: Converted parameters
        """
        # converted parameters
        converted = params.copy()

        # cumulative sum up to the previous parameter
        mass = 0

        # take mass of current bin to be fraction of mass assigned to the previous bins
        for p in self.params[1:-2]:
            converted[p] = params[p] * (1 - mass)
            mass += converted[p]

        # assign remaining mass to last bin
        converted[self.params[-2]] = 1 - mass

        return converted

    def to_fractional(self, params: Dict[str, float]) -> Dict[str, float]:
        """
        Invert the to_nominal operation: Convert representation of fractions which sum to 1
        back to representation of fraction of total mass to the left.

        :param params: Converted parameters (nominal space)
        :return: Original parameters (parameter space)
        """
        # converted parameters
        converted = params.copy()

        # cumulative sum up to the previous parameter
        mass = 0

        # Iterate over parameters
        for p in self.params[1:-2]:
            # Calculate original parameter value
            converted[p] = params[p] / (1 - mass)

            # update cumulative sum
            mass += params[p]

        # last parameter is simply what's left
        converted[self.params[-2]] = 1 - mass

        return converted

    def get_pdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get PDF.

        :param kwargs: Parameter values S1, S2, ..., Sk-1
        :return: Function that computes the PDF for given S values
        """

        # normalize and add boundary parameters
        values = self.to_nominal(kwargs | self.fixed_params)

        @Parametrization._accepts_scalars
        def pdf(S: np.ndarray) -> np.ndarray:
            """
            The PDF.

            :param S: Selection coefficients
            :return: Probability density
            """
            x = np.zeros_like(S, dtype=float)

            # iterate over parameters and assign values
            for i, p in enumerate(self.params):
                x[(self.intervals[i] < S) & (S <= self.intervals[i + 1])] = values[p] / self.interval_sizes[i]

            return x

        return pdf

    def get_cdf(self, **kwargs) -> Callable[[np.ndarray], np.ndarray]:
        """
        Get CDF.

        :param kwargs: Parameter values S1, S2, ..., Sk-1
        :return: Function that computes the CDF for given S values
        """

        # normalize and add boundary parameters
        values = self.to_nominal(kwargs | self.fixed_params)

        # convert to numpy arrays
        intervals = np.array(self.intervals)
        interval_sizes = np.array(self.interval_sizes)

        # convert to ordered array values
        vals = np.array([values[self.params[i]] for i in range(self.k)])

        @Parametrization._accepts_scalars
        def cdf(S: np.ndarray) -> np.ndarray:
            """
            The CDF.

            :param S: Selection coefficients
            :return: Cumulative probability
            """
            # iterate over parameters and assign values
            cum = np.cumsum([values[self.params[i]] for i in range(self.k)])

            # obtain bin indices
            i = np.sum(self.intervals[:, None] <= S[None, :], axis=0) - 1

            # make sure we don't go out of bounds which can happen if S is np.inf
            i[i >= self.k] = self.k - 1

            # cumulative probability up into previous bin
            cum_prev = cum[np.maximum(i - 1, np.zeros_like(S, dtype=int))]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)

                # probability in current bin
                cum_within = np.abs(intervals[i] - S) / interval_sizes[i] * vals[i]

            cum_within[np.isnan(cum_within)] = 0

            # return cumulative probability
            return cum_prev + cum_within

        return cdf



"""
fastDFE package.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-03-10"

__version__ = '1.1.13'

import logging
import sys

import jsonpickle
import numpy as np
import pandas as pd
from tqdm import tqdm

from .json_handlers import DataframeHandler, SpectrumHandler, SpectraHandler, NumpyArrayHandler
from .spectrum import Spectrum, Spectra

# register custom handles
jsonpickle.handlers.registry.register(pd.DataFrame, DataframeHandler)
jsonpickle.handlers.registry.register(Spectrum, SpectrumHandler)
jsonpickle.handlers.registry.register(Spectra, SpectraHandler)
jsonpickle.handlers.registry.register(np.ndarray, NumpyArrayHandler)


class TqdmLoggingHandler(logging.Handler):
    """
    TQDM logging handler
    """

    def __init__(self, level=logging.NOTSET):
        """
        Initialize the handler.

        :param level:
        """
        super().__init__(level)

    def emit(self, record):
        """
        Emit a record.
        """
        try:
            msg = self.format(record)

            # we write to stderr as the progress bar
            # to make the two work together
            tqdm.write(msg, file=sys.stderr)
            self.flush()
        except Exception:
            self.handleError(record)


class ColoredFormatter(logging.Formatter):
    """
    Colored formatter.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the formatter.
        """
        super().__init__(*args, **kwargs)

        self.colors = {
            "DEBUG": "\033[36m",  # Cyan
            "INFO": "\033[32m",  # Green
            "WARNING": "\033[33m",  # Yellow
            "ERROR": "\033[31m",  # Red
            "CRITICAL": "\033[31m",  # Red
        }

        self.reset = "\033[0m"

    def format(self, record):
        """
        Format the record.
        """
        color = self.colors.get(record.levelname, self.reset)

        formatted = super().format(record)

        # remove package name
        formatted = formatted.replace(record.name, record.name.split('.')[-1])

        return f"{color}{formatted}{self.reset}"


# configure logger
logger = logging.getLogger('fastdfe')

# don't propagate to the root logger
logger.propagate = False

# set to INFO by default
logger.setLevel(logging.INFO)

# let TQDM handle the logging
handler = TqdmLoggingHandler()

# define a Formatter with colors
formatter = ColoredFormatter('%(levelname)s:%(name)s: %(message)s')

handler.setFormatter(formatter)
logger.addHandler(handler)

# def raise_on_warning(message, category, filename, lineno, file=None, line=None):
#    """
#    Raise exception on warning.
#    """
#    raise Exception(warnings.formatwarning(message, category, filename, lineno, line))

# warnings.showwarning = raise_on_warning

# load class from modules
from .parametrization import Parametrization, GammaExpParametrization, DiscreteParametrization, \
    GammaDiscreteParametrization, DisplacedGammaParametrization, DiscreteFractionalParametrization
from .config import Config
from .settings import Settings
from .abstract_inference import Inference
from .base_inference import BaseInference, InferenceResults
from .joint_inference import JointInference, SharedParams
from .optimization import Covariate
from .simulation import Simulation
from .spectrum import Spectrum, Spectra
from .parser import Parser, Stratification, BaseTransitionStratification, BaseContextStratification, \
    DegeneracyStratification, TransitionTransversionStratification, AncestralBaseStratification, \
    SynonymyStratification, VEPStratification, SnpEffStratification, ContigStratification, ChunkedStratification, \
    RandomStratification, TargetSiteCounter
from .io_handlers import VCFHandler, FASTAHandler, GFFHandler, FileHandler
from .annotation import Annotator, Annotation, MaximumParsimonyAncestralAnnotation, SiteInfo, \
    MaximumLikelihoodAncestralAnnotation, DegeneracyAnnotation, SynonymyAnnotation, SubstitutionModel, \
    K2SubstitutionModel, JCSubstitutionModel, PolarizationPrior, KingmanPolarizationPrior, AdaptivePolarizationPrior
from .filtration import Filterer, Filtration, SNPFiltration, PolyAllelicFiltration, CodingSequenceFiltration, \
    SNVFiltration, DeviantOutgroupFiltration, AllFiltration, NoFiltration, BiasedGCConversionFiltration, \
    ExistingOutgroupFiltration, ContigFiltration

__all__ = [
    'Parametrization',
    'GammaExpParametrization',
    'DiscreteParametrization',
    'GammaDiscreteParametrization',
    'DisplacedGammaParametrization',
    'DiscreteFractionalParametrization',
    'Config',
    'Settings',
    'Inference',
    'BaseInference',
    'JointInference',
    'Simulation',
    'SharedParams',
    'Covariate',
    'Spectrum',
    'Spectra',
    'Parser',
    'FileHandler',
    'VCFHandler',
    'FASTAHandler',
    'GFFHandler',
    'TargetSiteCounter',
    'Stratification',
    'BaseTransitionStratification',
    'BaseContextStratification',
    'DegeneracyStratification',
    'TransitionTransversionStratification',
    'AncestralBaseStratification',
    'SynonymyStratification',
    'VEPStratification',
    'SnpEffStratification',
    'ContigStratification',
    'ChunkedStratification',
    'RandomStratification',
    'Annotator',
    'Annotation',
    'SiteInfo',
    'MaximumParsimonyAncestralAnnotation',
    'MaximumLikelihoodAncestralAnnotation',
    'PolarizationPrior',
    'KingmanPolarizationPrior',
    'AdaptivePolarizationPrior',
    'SubstitutionModel',
    'K2SubstitutionModel',
    'JCSubstitutionModel',
    'DegeneracyAnnotation',
    'SynonymyAnnotation',
    'Filtration',
    'CodingSequenceFiltration',
    'DeviantOutgroupFiltration',
    'AllFiltration',
    'NoFiltration',
    'Filterer',
    'SNPFiltration',
    'SNVFiltration',
    'PolyAllelicFiltration',
    'BiasedGCConversionFiltration',
    'ExistingOutgroupFiltration',
    'ContigFiltration',
    'Filterer',
]



"""
Bootstrap utilities.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import logging
from typing import Literal, Sequence

import numpy as np
from scipy.stats import norm as normal

# get logger
logger = logging.getLogger('fastdfe').getChild('Bootstrap')


class Bootstrap:
    """
    Bootstrap utilities.
    """

    @staticmethod
    def get_bounds_from_quantile(data: Sequence, a1: float, a2: float, n: int) -> (float, float):
        """
        Get confidence interval bounds.

        :param data: Sorted data
        :param a1: Lower quantile
        :param a2: Upper quantile
        :param n: Number of data points
        :return: lower bound and upper bound
        """
        if np.isnan(a1) or np.isnan(a2):
            return [None, None]

        return data[max(round(a1 * n), 0)], data[min(round(a2 * n), n - 1)]

    @staticmethod
    def get_ci_percentile(bootstraps: Sequence, a: float) -> (float, float):
        """
        Get the (1 - a)% confidence intervals using the percentile bootstrap.

        :param bootstraps: List of bootstraps
        :param a: Confidence level
        :return: lower bound and upper bound
        """
        if not 0 <= a <= 0.5:
            raise ValueError('Confidence level must be between 0 and 0.5.')

        return Bootstrap.get_bounds_from_quantile(np.sort(bootstraps), a, 1 - a, len(bootstraps))

    @staticmethod
    def get_ci_bca(bootstraps: Sequence, original: float, a: float) -> (float, float):
        """
        Get the (1 - a)% confidence intervals using the BCa method.
        cf. An Introduction to the Bootstrap, Bradley Efron, Robert J. Tibshirani, section 14.2.

        :param bootstraps: List of bootstraps
        :param original: Original value
        :param a: Confidence level
        :return: lower bound and upper bound
        """
        if len(bootstraps) == 0:
            return [0, 0]

        n = len(bootstraps)
        data = np.sort(np.array(bootstraps))

        theta_hat_i = np.array([np.var(np.delete(data, i)) for i in range(n)])
        theta_hat = sum(theta_hat_i) / n

        a_hat = sum((theta_hat - theta_hat_i) ** 3) / (6 * (sum((theta_hat - theta_hat_i) ** 2)) ** (3 / 2))

        # we add epsilon here to avoid getting -inf when sum(data < original) / n is 0
        z0_hat = normal.ppf(sum(data < original) / n + np.finfo(float).eps)
        z_a = normal.ppf(a)

        a1 = normal.cdf(z0_hat + (z0_hat + z_a) / (1 - a_hat * (z0_hat + z_a)))
        a2 = normal.cdf(z0_hat + (z0_hat - z_a) / (1 - a_hat * (z0_hat - z_a)))

        return Bootstrap.get_bounds_from_quantile(data, a1, a2, n)

    @staticmethod
    def get_errors(
            values: Sequence,
            bs: np.ndarray,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile'
    ) -> (np.ndarray, np.ndarray):
        """
        Get error values and confidence intervals from the list of original
        values and its bootstraps for a list of parameters.

        :param values: The original values.
        :param bs: Array containing bootstraps as first and the values as second dimension.
        :param ci_level: The confidence level.
        :param bootstrap_type: The bootstrap type.
        :return: Arrays of errors and confidence intervals.
        """
        # number of values
        n_values = len(values)

        # replace value by mean
        means = np.mean(bs, axis=0)

        # determine confidence intervals and errors on y axis
        if bootstrap_type == 'percentile':
            cis = np.array([Bootstrap.get_ci_percentile(bs[:, i], ci_level) for i in range(n_values)]).T

            # Determine errors using mean values of the bootstraps.
            # Note that this sometimes causes the errors to be negative.
            errors = np.array([means - cis[0], cis[1] - means])
        elif bootstrap_type == 'bca':
            cis = np.array([Bootstrap.get_ci_bca(bs[:, i], values[i], ci_level) for i in range(n_values)]).T

            # raise warning if some confidence intervals could not be determined
            if np.sum(np.equal(cis, None)) > 0:
                logger.warning('Some confidence intervals could not be computed.')

                # set undefined confidence intervals to values
                mask = np.equal(cis, None)
                for j in range(cis.shape[1]):
                    if mask[:, j].any():
                        cis[mask[:, j], j] = values[j]

            # determine error using original values
            errors = np.array([values - cis[0], cis[1] - values])
        else:
            raise NotImplementedError(f"Bootstrap type {bootstrap_type} not supported.")

        # issue notice if some errors are negative
        if np.sum(np.less(errors, 0)) > 0:
            logger.debug('Some computed errors were negative and were adjusted to 0.')

            # set negative errors to 0
            errors[errors < 0] = 0

        return errors, cis



"""
Abstract inference class and static utilities.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-03-12"

import logging
from abc import ABC, abstractmethod
from typing import List, Optional, Literal, Tuple, Dict, Sequence

import jsonpickle
import numpy as np
import pandas as pd
from typing_extensions import Self

from .bootstrap import Bootstrap
from .parametrization import Parametrization, _from_string

logger = logging.getLogger("fastdfe")


class Inference:
    """
    Static utility methods for inference objects.
    """

    @staticmethod
    def plot_discretized(
            inferences: List['AbstractInference'],
            intervals: Sequence = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            file: str = None,
            show: bool = True,
            title: str = 'discretized DFEs',
            labels: Sequence = None,
            ax: 'plt.Axes' = None,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            **kwargs

    ) -> 'plt.Axes':
        """
        Visualize several discretized DFEs given by the list of inference objects.

        :param inferences: List of inference objects.
        :param intervals: Intervals over ``(-inf, inf)`` to use for discretization.
        :param confidence_intervals: Whether to plot confidence intervals.
        :param ci_level: Confidence level for confidence intervals.
        :param bootstrap_type: Type of bootstrap to use for confidence intervals.
        :param file: Path to file to save the plot to.
        :param show: Whether to show the plot.
        :param title: Title of the plot.
        :param labels: Labels for the DFEs.
        :param kwargs: Additional arguments for the plot.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes of the plot.
        """
        from .visualization import Visualization

        # get data from inference objects
        values = []
        errors = []
        for i, inference in enumerate(inferences):
            val, errs = inference.get_discretized(
                intervals=np.array(intervals),
                confidence_intervals=confidence_intervals,
                ci_level=ci_level,
                bootstrap_type=bootstrap_type
            )

            values.append(val)
            errors.append(errs)

        # plot DFEs
        return Visualization.plot_discretized(
            values=values,
            errors=errors,
            labels=labels,
            file=file,
            show=show,
            intervals=np.array(intervals),
            title=title,
            ax=ax,
            kwargs_legend=kwargs_legend
        )

    @staticmethod
    def plot_continuous(
            inferences: List['AbstractInference'],
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            file: str = None,
            show: bool = True,
            title: str = 'continuous DFEs',
            labels: Sequence = None,
            scale: Literal['lin', 'log', 'symlog'] = 'lin',
            scale_density: bool = False,
            ax: 'plt.Axes' = None,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            **kwargs

    ) -> 'plt.Axes':
        """
        Visualize several DFEs given by the list of inference objects.
        By default, the PDF is plotted as is. Due to the logarithmic scale on
        the x-axis, we may get a wrong intuition on how the mass is distributed,
        however. To get a better intuition, we can optionally scale the density
        by the x-axis interval size using ``scale_density = True``. This has the
        disadvantage that the density now changes for x, so that even a constant
        density will look warped.

        :param inferences: List of inference objects.
        :param intervals: Intervals to use for discretization.
        :param confidence_intervals: Whether to plot confidence intervals.
        :param ci_level: Confidence level for confidence intervals.
        :param bootstrap_type: Type of bootstrap to use for confidence intervals.
        :param file: Path to file to save the plot to.
        :param show: Whether to show the plot.
        :param title: Title of the plot.
        :param labels: Labels for the DFEs.
        :param scale: y-scale of the plot.
        :param scale_density: Whether to scale the density by the x-axis interval size.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :param kwargs: Additional arguments for the plot.
        :return: Axes of the plot.
        """
        from .visualization import Visualization

        # get data from inference objects
        values = []
        errors = []
        for i, inf in enumerate(inferences):
            val, errs = inf.get_discretized(
                intervals=intervals,
                confidence_intervals=confidence_intervals,
                ci_level=ci_level,
                bootstrap_type=bootstrap_type
            )

            values.append(val)
            errors.append(errs)

        # plot DFEs
        return Visualization.plot_continuous(
            bins=intervals,
            **locals()
        )

    @staticmethod
    def plot_inferred_parameters(
            inferences: List['AbstractInference'],
            labels: Sequence,
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            file: str = None,
            show: bool = True,
            title: str = 'parameter estimates',
            scale: Literal['lin', 'log', 'symlog'] = 'log',
            ax: 'plt.Axes' = None,
            kwargs_legend: dict = dict(prop=dict(size=8), loc='upper right'),
            **kwargs

    ) -> 'plt.Axes':
        """
        Visualize several discretized DFEs given by the list of inference objects.
        Note that the DFE parametrization needs to be the same for all inference objects.

        :param inferences: List of inference objects.
        :param labels: Unique labels for the DFEs.
        :param scale: y-scale of the plot.
        :param confidence_intervals: Whether to plot confidence intervals.
        :param ci_level: Confidence level for confidence intervals.
        :param bootstrap_type: Type of bootstrap to use for confidence intervals.
        :param file: Path to file to save the plot to.
        :param show: Whether to show the plot.
        :param title: Title of the plot.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :return: Axes of the plot.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :param kwargs: Additional arguments which are ignored.
        :raises ValueError: If no inference objects are given.
        """
        from .visualization import Visualization

        if len(inferences) == 0:
            raise ValueError('No inference objects given.')

        # get sorted list of parameter names
        param_names = sorted(list(inferences[0].get_bootstrap_param_names()))

        # get errors and values
        errors, values = Inference.get_errors_params_mle(
            bootstrap_type=bootstrap_type,
            ci_level=ci_level,
            confidence_intervals=confidence_intervals,
            inferences=inferences,
            labels=labels,
            param_names=param_names
        )

        return Visualization.plot_inferred_parameters(
            values=values,
            errors=errors,
            param_names=param_names,
            file=file,
            show=show,
            title=title,
            labels=labels,
            scale=scale,
            legend=len(labels) > 1,
            kwargs_legend=kwargs_legend,
            ax=ax,
        )

    @staticmethod
    def plot_inferred_parameters_boxplot(
            inferences: List['AbstractInference'],
            labels: Sequence,
            file: str = None,
            show: bool = True,
            title: str = 'parameter estimates',
            **kwargs
    ) -> 'plt.Axes':
        """
        Visualize several discretized DFEs given by the list of inference objects.
        Note that the DFE parametrization needs to be the same for all inference objects.

        :param inferences: List of inference objects.
        :param labels: Unique labels for the DFEs.
        :param file: Path to file to save the plot to.
        :param show: Whether to show the plot.
        :param title: Title of the plot.
        :param kwargs: Additional arguments for the plot.
        :return: Axes of the plot.
        :raises ValueError: If no inference objects are given or no bootstraps are found.
        """
        from .visualization import Visualization

        if len(inferences) == 0:
            raise ValueError('No inference objects given.')

        # get sorted list of parameter names
        param_names = sorted(list(inferences[0].get_bootstrap_param_names()))

        if inferences[0].bootstraps is None:
            raise ValueError('No bootstraps found.')

        # create dict of dataframes
        values = dict((k, inf.bootstraps) for k, inf in zip(labels, inferences))

        return Visualization.plot_inferred_parameters_boxplot(
            values=values,
            param_names=param_names,
            file=file,
            show=show,
            title=title,
        )

    @staticmethod
    def get_cis_params_mle(
            bootstrap_type: Literal['percentile', 'bca'],
            ci_level: float,
            inferences: List['AbstractInference'],
            labels: Sequence,
            param_names: Sequence
    ) -> dict[str, Optional[Dict[str, Tuple[float, float]]]]:
        """
        Get confidence intervals for the MLE parameters.

        :param bootstrap_type: Type of bootstrap to use for confidence intervals.
        :param ci_level: Confidence level for confidence intervals.
        :param inferences: List of inference objects.
        :param labels: Labels for the DFEs.
        :param param_names: Names of the parameters.
        :return: Dictionary of confidence intervals indexed by labels.
        """
        cis = {}
        for label, inf in zip(labels, inferences):

            if inf.bootstraps is not None:

                values = list(inf.get_bootstrap_params()[k] for k in param_names)

                res = Bootstrap.get_errors(
                    values=values,
                    bs=inf.bootstraps[param_names].to_numpy(),
                    bootstrap_type=bootstrap_type,
                    ci_level=ci_level
                )

                cis[label] = {k: tuple(res[1][:, i]) for i, k in enumerate(param_names)}
            else:
                cis[label] = None

        return cis

    @staticmethod
    def get_errors_params_mle(
            bootstrap_type: Literal['percentile', 'bca'],
            ci_level: float,
            confidence_intervals: bool,
            inferences: List['AbstractInference'],
            labels: Sequence,
            param_names: Sequence
    ) -> (Dict[str, Tuple[np.ndarray, np.ndarray] | None], Dict[str, np.ndarray]):
        """
        Get errors and values for MLE params of inferences.

        :param bootstrap_type: Type of bootstrap to use.
        :param ci_level: Confidence level for confidence intervals.
        :param confidence_intervals: Whether to compute confidence intervals.
        :param inferences: List of inference objects.
        :param labels: Labels for the inferences.
        :param param_names: Names of the parameters to get errors and values for.
        :return: dictionary of errors and dictionary of values
        """
        errors = {}
        values = {}
        for label, inf in zip(labels, inferences):

            values[label] = list(inf.get_bootstrap_params()[k] for k in param_names)

            # whether to compute errors
            if confidence_intervals and inf.bootstraps is not None:

                # use mean of bootstraps instead of original values
                if bootstrap_type == 'percentile':
                    values[label] = inf.bootstraps[param_names].mean().to_list()

                # compute errors
                errors[label], _ = Bootstrap.get_errors(
                    values=values[label],
                    bs=inf.bootstraps[param_names].to_numpy(),
                    bootstrap_type=bootstrap_type,
                    ci_level=ci_level
                )
            else:
                errors[label] = None

        return errors, values

    @staticmethod
    def get_discretized(
            inferences: List['AbstractInference'],
            labels: Sequence,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile'

    ) -> (np.ndarray, Optional[np.ndarray]):
        """
        Get values and errors of discretized DFE.

        :param inferences: List of inference objects.
        :param labels: Labels for the DFEs.
        :param bootstrap_type: Type of bootstrap to use
        :param ci_level: Confidence interval level
        :param confidence_intervals: Whether to compute confidence intervals
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bars.
        :return: Array of values and array of errors
        """
        values = {}
        errors = {}

        for label, inf in zip(labels, inferences):

            if confidence_intervals and inf.bootstraps is not None:
                # get bootstraps and errors if specified
                errs, _, bs, means, vals = Inference.get_stats_discretized(
                    params=inf.get_bootstrap_params(),
                    bootstraps=inf.bootstraps,
                    model=inf.model,
                    ci_level=ci_level,
                    intervals=intervals,
                    bootstrap_type=bootstrap_type
                )
            else:
                # otherwise just get discretized values
                vals = Inference.compute_histogram(
                    params=inf.get_bootstrap_params(),
                    model=inf.model,
                    intervals=intervals
                )
                errs, means, bs = None, None, None

            # whether to use the mean of all bootstraps instead of the original values
            use_means = confidence_intervals and inf.bootstraps is not None and bootstrap_type == 'percentile'

            if use_means:
                vals = np.mean(bs, axis=0)

            values[label] = vals
            errors[label] = errs

        return values, errors

    @staticmethod
    def get_stats_discretized(
            params: dict,
            bootstraps: pd.DataFrame,
            model: Parametrization | str,
            ci_level: float = 0.05,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile'

    ) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
        """
        Compute errors and confidence interval for a discretized DFE.

        :param params: Parameters of the model
        :param bootstraps: Bootstrapped samples
        :param model: DFE parametrization
        :param ci_level: Confidence interval level
        :param intervals: Array of interval boundaries yielding ``intervals.shape[0] - 1`` bins.
        :param bootstrap_type: Type of bootstrap
        :return: Arrays of errors, confidence intervals, bootstraps, means and values
        """
        # discretize MLE DFE
        values = Inference.compute_histogram(model, params, intervals)

        # calculate bootstrapped histograms
        # get discretized DFE per bootstrap sample
        bs = np.array([Inference.compute_histogram(model, dict(r), intervals) for _, r in bootstraps.iterrows()])

        errors, cis = Bootstrap.get_errors(
            values=values,
            bs=bs,
            bootstrap_type=bootstrap_type,
            ci_level=ci_level
        )

        # calculate mean values
        means = np.mean(bs, axis=0)

        return errors, cis, bs, means, values

    @staticmethod
    def compute_histogram(
            model: Parametrization | str,
            params: dict,
            intervals: np.ndarray
    ) -> np.ndarray:
        """
        Discretize the DFE given a DFE parametrization and its parameter values.

        :param model: DFE parametrization
        :param params: Parameters of the model
        :param intervals: Array of interval boundaries yielding ``intervals.shape[0] - 1`` bins.
        :return: Discretized DFE
        """
        # discrete DFE
        y = _from_string(model)._discretize(params, intervals)

        # return normalized histogram
        return y / y.sum()


class AbstractInference(ABC):
    """
    Base class for main Inference and polyDFE wrapper.
    """

    def __init__(self, **kwargs):
        """
        Initialize the inference.

        :param kwargs: Keyword arguments
        """
        self._logger = logger.getChild(self.__class__.__name__)

        self.bootstraps: Optional[pd.DataFrame] = None
        self.params_mle: Optional[dict] = None
        self.model: Optional[Parametrization] = None

    @abstractmethod
    def get_bootstrap_params(self) -> Dict[str, float]:
        """
        Get the parameters to be included in the bootstraps.

        :return: Parameters to be included in the bootstraps
        """
        pass

    def get_discretized(
            self,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile'
    ) -> (np.ndarray, np.ndarray):
        """
        Get discretized DFE.

        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence interval level
        :param confidence_intervals: Whether to return confidence intervals
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bins.
        :return: Array of values and array of deviations
        """
        values, errors = Inference.get_discretized(
            inferences=[self],
            labels=['all'],
            intervals=intervals,
            confidence_intervals=confidence_intervals,
            ci_level=ci_level,
            bootstrap_type=bootstrap_type
        )

        return values['all'], errors['all']

    def plot_discretized(
            self,
            file: str = None,
            show: bool = True,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            title: str = 'discretized DFE',
            ax: 'plt.Axes' = None,
            kwargs_legend: dict = dict(prop=dict(size=8)),
    ) -> 'plt.Axes':
        """
        Plot discretized DFE.

        :param title: Title of the plot
        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence interval level
        :param confidence_intervals: Whether to plot confidence intervals
        :param file: File to save the plot to
        :param show: Whether to show the plot
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bars.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes
        """
        return Inference.plot_discretized(
            inferences=[self],
            file=file,
            show=show,
            intervals=intervals,
            confidence_intervals=confidence_intervals,
            ci_level=ci_level,
            bootstrap_type=bootstrap_type,
            title=title,
            kwargs_legend=kwargs_legend,
            ax=ax
        )

    def get_cis_params_mle(
            self,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            ci_level: float = 0.05,
            param_names: Optional[list[str]] = None
    ) -> Dict[str, Tuple[float, float]]:
        """
        Get confidence intervals for the MLE parameters.

        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence interval level
        :param param_names: Names of the parameters to return confidence intervals for
        :return: Confidence intervals for the parameters
        """
        if param_names is None:
            param_names = self.get_bootstrap_param_names()

        return Inference.get_cis_params_mle(
            inferences=[self],
            bootstrap_type=bootstrap_type,
            ci_level=ci_level,
            param_names=param_names,
            labels=['all']
        )['all']

    def to_json(self) -> str:
        """
        Serialize object.

        :return: JSON string
        """
        return jsonpickle.encode(self, indent=4, warn=True)

    def to_file(self, file: str):
        """
        Save object to file.

        :param file: File to save to
        """
        with open(file, 'w') as fh:
            fh.write(self.to_json())

    @classmethod
    def from_json(cls, json: str, classes=None) -> Self:
        """
        Unserialize object.

        :param classes: Classes to be used for unserialization
        :param json: JSON string
        """
        return jsonpickle.decode(json, classes=classes)

    @classmethod
    def from_file(cls, file: str, classes=None) -> Self:
        """
        Load object from file.

        :param classes: Classes to be used for unserialization
        :param file: File to load from
        """
        with open(file, 'r') as fh:
            return cls.from_json(fh.read(), classes)

    @abstractmethod
    def get_bootstrap_param_names(self) -> List[str]:
        """
        Get the names of the parameters to be included in the bootstraps.
        """
        pass



"""
polyDFE wrapper utilities.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import copy
import json
import logging
import subprocess
import tempfile
import time
from typing import Callable, List, Optional, Literal, Dict

import numpy as np
import pandas as pd

from typing_extensions import Self

from .abstract_inference import AbstractInference, Inference
from .config import Config
from .parametrization import _from_string, Parametrization, DiscreteParametrization
from .polydfe_utils import create_sfs_config, models

# get logger
logger = logging.getLogger('fastdfe').getChild('PolydfeWrapper')


class PolyDFEResult:
    """
    Class for parsing polyDFE output.
    """
    # default postprocessing source
    default_postprocessing_source = '../resources/polydfe/postprocessing/script.R'

    # mapping of polyDFE params to new name
    param_mapping = dict(
        eps_an='eps'
    )

    def __init__(
            self,
            output_file: str,
            additional_data: dict = {},
            postprocessing_source: str = default_postprocessing_source
    ):
        """
        Create summary object.

        :param output_file: Path to polyDFE output file
        :param additional_data: Additional data to add to summary
        :param postprocessing_source: Path to polyDFE postprocessing script
        """
        self.output_file = output_file
        self.postprocessing_source = postprocessing_source

        self.data = self.parse_output() | additional_data

    @staticmethod
    def map_name(name: str) -> str:
        """
        Map polyDFE parameter name to new name.

        :param name: polyDFE parameter name
        :return: New name
        """
        if name in PolyDFEResult.param_mapping:
            return PolyDFEResult.param_mapping[name]

        return name

    def parse_output(self) -> dict:
        """
        Parse output from polyDFE output file.

        :return: Dictionary of parsed data
        """
        from rpy2.robjects.packages import importr

        # use the polyDFE R postprocessing script to parse the output file
        output = self.get_postprocessing_wrapper().parseOutput(self.output_file)[0]

        # convert to JSON
        rj = importr('RJSONIO')
        json_str = rj.toJSON(output)[0]

        # load dict from JSON string
        data = json.loads(json_str)

        return dict(
            output_file=data['input'],
            model=data['model'],
            likelihood=data['lk'],
            criteria=data['criteria'],
            params_mle=dict(
                # non-nuisance parameters
                # map polyDFE param name to new name
                (PolyDFEResult.map_name(k), v) for k, v in data['values'][0].items() if not k.startswith('r ')
            ) | dict(
                # nuisance parameters
                r=[data['values'][0][f"r {i}"] for i in range(2, data['n'])]
            ),
            n=data['n'],
            expec=data['expec'][0],
            alpha=self.get_alpha(output)
        )

    def get_alpha(self, parsed_output: 'rpy2.robjects.vectors.ListVector') -> float:
        """
        Get alpha, the proportion of beneficial non-synonymous substitutions.

        :return: Parsed output from polyDFE
        """
        return self.get_postprocessing_wrapper().estimateAlpha(parsed_output)[0]

    def get_postprocessing_wrapper(self) -> 'ro.R':
        """
        Get the wrapped polyDFE postprocessing source.

        :return: R object
        """
        import rpy2.robjects as ro

        ps = ro.r
        ps.source(self.postprocessing_source)

        return ps

    def to_json(self) -> str:
        """
        Convert object to JSON.

        :return: JSON string
        """
        return json.dumps(self.data, indent=4)

    def to_file(self, file: str):
        """
        Save object to file.

        :return: File path to save to
        """
        with open(file, 'w') as fh:
            fh.write(self.to_json())


class PolyDFE(AbstractInference):
    """
    polyDFE wrapper.

    Note that this class has dependencies outside of pip.
    Currently only model C is fully implemented.
    """

    def __init__(self, config: Config):
        """
        Create polyDFE wrapper.

        :param config: Config object
        """
        super().__init__()

        self.config = config

        # polyDFE output file
        self.output_file: Optional[str] = None

        # the total execution time in seconds
        self.execution_time: float = 0

        # polyDFE summary
        self.summary: Optional[PolyDFEResult] = None

        # bootstrap samples
        self.bootstraps: Optional[pd.DataFrame] = None

    @staticmethod
    def map_polydfe_model(model: str) -> str:
        """
        Map polyDFE model name to native equivalent.

        :param model: polyDFE model name
        :return: Native model name
        """
        return {v: k for k, v in models.items()}[model]

    @classmethod
    def from_config(cls, config: Config) -> Self:
        """
        Load from config object.

        :param config: Config object
        """
        return cls(config)

    @classmethod
    def from_config_file(cls, file: str) -> Self:
        """
        Load from config object.

        :param file: Config file path, possibly URL.
        :return: Inference object.
        """
        return cls.from_config(Config.from_file(file))

    def run(
            self,
            output_file: str,
            binary: str = 'polydfe',
            wd: str = None,
            execute: Callable = None,
            postprocessing_source: str = PolyDFEResult.default_postprocessing_source
    ) -> PolyDFEResult:
        """
        Run polyDFE.

        :param postprocessing_source: polyDFE postprocessing source
        :param execute: Function for executing shell commands
        :param wd: Working directory
        :param binary: polyDFE binary
        :param output_file: Output file
        :return: polyDFE summary
        """
        start_time = time.time()

        # define default function for executing command
        if execute is None:
            def shell(command: str):
                """
                Execute shell command.

                :param command: Command string
                """
                return subprocess.run(command, check=True, cwd=wd, shell=True)

            execute = shell

        # save the spectra and init file, so they can be reviewed if necessary
        # use temp file to generate the file name.
        with open(tempfile.NamedTemporaryFile().name, 'w') as spectra_file:
            with open(tempfile.NamedTemporaryFile().name, 'w') as init_file:
                # save files
                self.config.create_polydfe_sfs_config(spectra_file.name)
                self.config.create_polydfe_init_file(init_file.name, n=self.config.data['sfs_neut'].n)

                # add number of fragment if model is DiscreteParametrization
                model = _from_string(self.config.data['model'])
                k = str(model.k - 1) + ' ' if isinstance(model, DiscreteParametrization) else ''

                # construct command string
                command = (f"{binary} "
                           f"-d {spectra_file.name} "
                           f"-m {self.config.get_polydfe_model()} {k}"
                           f"-i {init_file.name} 1 "
                           f"-v 1 > {output_file}")

                # log command signature
                self._logger.info(f"Running: '{command}'")

                # execute command
                execute(command)

        # add execution time
        self.execution_time += time.time() - start_time

        # create summary from output file
        self.summary = PolyDFEResult(
            output_file=output_file,
            additional_data=dict(execution_time=self.execution_time),
            postprocessing_source=postprocessing_source
        )

        return self.summary

    def create_bootstrap(self) -> Config:
        """
        Create a bootstrap sample using polyDFE's resampling.

        :return: Config object
        """
        if self.summary is None:
            raise Exception('PolyDFE needs to be run before creating bootstrap samples.')

        # postprocessing wrapper
        ps = self.summary.get_postprocessing_wrapper()

        # create temporary SFS config file
        with tempfile.NamedTemporaryFile() as tmp:
            create_sfs_config(
                file=tmp.name,
                sfs_neut=self.config.data['sfs_neut']['all'],
                sfs_sel=self.config.data['sfs_sel']['all']
            )

            # create bootstrap SFS config
            ps.bootstrapData(tmp.name, rep=1)

            # file name of resampled SFS config
            sfs_config_file = tmp.name + '_1'

            # use exiting config as template and load init and SFS file
            config = copy.deepcopy(self.config)
            config.parse_polydfe_sfs_config(sfs_config_file)
            config.data['x0'] = dict(all=self.summary.data['params_mle'])

        return config

    def add_bootstraps(self, bootstraps: 'List[PolyDFE]'):
        """
        Add bootstraps samples.

        :param bootstraps: List of bootstrap samples
        """
        # load MLE params into dataframe
        self.bootstraps = pd.DataFrame([bs.get_bootstrap_params() for bs in bootstraps])

        # drop nuisance parameters and eps_cont
        self.bootstraps.drop(columns=['r', 'eps_cont'], inplace=True, errors='ignore')

        # update execution time
        self.execution_time += float(np.sum([bs.summary.data['execution_time'] for bs in bootstraps]))

    def plot_all(self, show: bool = True):
        """
        Plot a bunch of plots.

        :param show: Whether to show the plot
        """
        self.plot_inferred_parameters(show=show)
        self.plot_discretized(show=show)

    def plot_inferred_parameters(
            self,
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            file: str = None,
            show: bool = True,
            title: str = 'parameter estimates',
            scale: Literal['lin', 'log', 'symlog'] = 'log',
            legend: bool = True,
            ax: 'plt.Axes' = None,
            kwargs_legend: dict = dict(prop=dict(size=8), loc='upper right'),
    ) -> 'plt.Axes':
        """
        Visualize the inferred parameters and their confidence intervals.

        :param scale: Scale of the y-axis
        :param confidence_intervals: Whether to show confidence intervals
        :param ci_level: Confidence interval level
        :param bootstrap_type: Type of bootstrap
        :param legend: Show legend
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param title: Title of the plot
        :param show: Show the plot
        :param file: File to save the plot to
        :param show: Show the plot
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes object
        """
        return Inference.plot_inferred_parameters(
            inferences=[self],
            labels=['all'],
            confidence_intervals=confidence_intervals,
            ci_level=ci_level,
            bootstrap_type=bootstrap_type,
            file=file,
            show=show,
            title=title,
            legend=legend,
            scale=scale,
            ax=ax,
            kwargs_legend=kwargs_legend
        )

    def get_bootstrap_param_names(self) -> List[str]:
        """
        Parameter names for parameters included in bootstraps.

        :return: List of parameter names
        """
        return _from_string(self.config.data['model']).param_names + ['eps', 'alpha']

    def get_bootstrap_params(self) -> Dict[str, float]:
        """
        Get the parameters to be included in the bootstraps.

        :return: Dict of parameters
        """
        params = self.summary.data['params_mle'] | dict(alpha=self.summary.data['alpha'])

        # filter params
        return dict((k, params[k]) for k in self.get_bootstrap_param_names())

    @property
    def params_mle(self) -> dict:
        """
        Get the maximum likelihood estimate of the parameters.

        :return: Dict of parameters
        """
        return self.summary.data['params_mle']

    @params_mle.setter
    def params_mle(self, value):
        """
        Set the maximum likelihood estimate of the parameters.

        :param value: Dict of parameters
        """
        pass

    @property
    def model(self) -> Parametrization:
        """
        Get the DFE parametrization.

        :return: Parametrization
        """

        return _from_string(self.config.data['model'])

    @model.setter
    def model(self, value):
        """
        Set the DFE parametrization.

        :param value: Parametrization
        """
        pass



"""
Configuration class.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import json
import logging
from typing import List, Literal, Dict, Tuple

import yaml

from .io_handlers import download_if_url
from .json_handlers import CustomEncoder
from .optimization import Covariate
from .optimization import SharedParams, merge_dicts
from .parametrization import Parametrization, _from_string, _to_string
from .polydfe_utils import create_sfs_config, parse_init_file, create_init_file, models
from .spectrum import Spectra, parse_polydfe_sfs_config, Spectrum

logger = logging.getLogger('fastdfe').getChild('Config')


class Config:
    """
    Configuration class to be used for :class:`~fastdfe.base_inference.BaseInference` and
    :class:`~fastdfe.joint_inference.JointInference`.
    """

    def __init__(
            self,
            polydfe_spectra_config: str = None,
            polydfe_init_file: str = None,
            polydfe_init_file_id: int = 1,
            sfs_neut: Spectra | Spectrum = None,
            sfs_sel: Spectra | Spectrum = None,
            intervals_del: Tuple[float, float, int] = (-1.0e+8, -1.0e-5, 1000),
            intervals_ben: Tuple[float, float, int] = (1.0e-5, 1.0e4, 1000),
            integration_mode: Literal['midpoint', 'quad'] = 'midpoint',
            linearized: bool = True,
            model: Parametrization | str = 'GammaExpParametrization',
            seed: int = 0,
            x0: Dict[str, Dict[str, float]] = {},
            bounds: Dict[str, Tuple[float, float]] = {},
            scales: Dict[str, Literal['lin', 'log', 'symlog']] = {},
            loss_type: Literal['likelihood', 'L2'] = 'likelihood',
            opts_mle: dict = {},
            method_mle: str = 'L-BFGS-B',
            n_runs: int = 10,
            fixed_params: Dict[str, Dict[str, float]] = {},
            shared_params: List[SharedParams] = [],
            covariates: List[Covariate] = [],
            do_bootstrap: bool = False,
            n_bootstraps: int = 100,
            n_bootstrap_retries: int = 2,
            parallelize: bool = True,
            **kwargs
    ):
        """
        Create config object.

        :param polydfe_spectra_config: Path to polyDFE SFS config file.
        :param polydfe_init_file: Path to polyDFE init file.
        :param polydfe_init_file_id: ID of polyDFE init file.
        :param sfs_neut: Neutral SFS. Note that we require monomorphic counts to be specified in order to infer
            the mutation rate.
        :param sfs_sel: Selected SFS. Note that we require monomorphic counts to be specified in order to infer
            the mutation rate.
        :param intervals_del: ``(start, stop, n_interval)`` for deleterious population-scaled
            selection coefficients. The intervals will be log10-spaced.
        :param intervals_ben: Same as ``intervals_del`` but for positive selection coefficients.
        :param integration_mode: Integration mode, ``quad`` not recommended
        :param linearized: Whether to use the linearized version of the DFE, ``False`` not recommended.
        :param model: Parametrization of the DFE.
        :param seed: Seed for the random number generator. Use ``None`` for no seed.
        :param x0: Dictionary of initial values in the form ``{type: {param: value}}``
        :param bounds: Bounds for the optimization in the form {param: (lower, upper)}
        :param scales: Scales for the optimization in the form {param: scale}
        :param loss_type: Loss function to use.
        :param opts_mle: Options for the optimization.
        :param method_mle: Method to use for optimization. See `scipy.optimize.minimize` for available methods.
        :param n_runs: Number of independent optimization runs out of which the best one is chosen. The first run
            will use the initial values if specified. Consider increasing this number if the optimization does not
            produce good results.
        :param fixed_params: Fixed parameters for the optimization.
        :param shared_params: Shared parameters for the optimization.
        :param covariates: Covariates for the optimization.
        :param do_bootstrap: Whether to do bootstrapping automatically.
        :param n_bootstraps: Number of bootstraps.
        :param n_bootstrap_retries: Number of retries for bootstraps that did not terminate normally.
        :param parallelize: Whether to parallelize the optimization.
        :param kwargs: Additional keyword arguments which are ignored.
        """

        # save options
        self.data = dict(
            model=model,
            intervals_del=intervals_del,
            intervals_ben=intervals_ben,
            integration_mode=integration_mode,
            linearized=linearized,
            seed=seed,
            opts_mle=opts_mle,
            method_mle=method_mle,
            n_runs=n_runs,
            x0=x0,
            bounds=bounds,
            scales=scales,
            fixed_params=fixed_params,
            shared_params=shared_params,
            covariates=covariates,
            sfs_neut=Spectra.from_spectrum(sfs_neut) if isinstance(sfs_neut, Spectrum) else sfs_neut,
            sfs_sel=Spectra.from_spectrum(sfs_sel) if isinstance(sfs_sel, Spectrum) else sfs_sel,
            do_bootstrap=do_bootstrap,
            n_bootstraps=n_bootstraps,
            n_bootstrap_retries=n_bootstrap_retries,
            parallelize=parallelize,
            loss_type=loss_type
        )

        # parse spectra file if specified
        if polydfe_spectra_config is not None:
            self.parse_polydfe_sfs_config(polydfe_spectra_config)

        # parse init file if specified
        if polydfe_init_file is not None:
            self.parse_polydfe_init_file(polydfe_init_file, polydfe_init_file_id)

    def update(self, **kwargs) -> 'Config':
        """
        Update config with given data.

        :param kwargs: Data to update.
        :return: Updated config.
        """
        # convert spectrum to spectra objects
        for key in ['sfs_neut', 'sfs_sel']:
            if key in kwargs and isinstance(kwargs[key], Spectrum):
                kwargs[key] = Spectra.from_spectrum(kwargs[key])

        self.data |= kwargs

        return self

    def parse_polydfe_init_file(self, file: str, id: int = 1, type='all'):
        """
        Parse polyDFE init file.
        This will define the initial parameters and
        which ones will be held fixed during the optimization.

        :param type: Type of parameters to parse for.
        :param id: ID of the init file.
        :param file: Path to the init file.
        """
        fixed_params, x0 = parse_init_file(_from_string(self.data['model']).param_names, file, id)

        # merge with existing config
        self.data['fixed_params'] = merge_dicts(self.data['fixed_params'], dict(all=fixed_params))
        self.data['x0'] |= {type: x0}

    def create_polydfe_init_file(self, file: str, n: int, type: str = 'all'):
        """
        Create an init file for polyDFE.

        :param type: Type to use for the init file.
        :param n: SFS samples size.
        :param file: Path to the init file to be created.
        """
        create_init_file(
            file=file,
            fixed_params=list(self.data['fixed_params'][type].keys()) if type in self.data['fixed_params'] else [],
            x0=self.data['x0'][type] if type in self.data['x0'] else {},
            dfe_params=_from_string(self.data['model']).param_names,
            n=n
        )

    def parse_polydfe_sfs_config(self, file: str):
        """
        Parse frequency spectra and mutational target site from
        polyDFE configuration file.

        :param file: Path to the polyDFE config file.
        """
        spectra = parse_polydfe_sfs_config(file)

        # merge into data dictionary
        self.data |= dict(
            sfs_neut=Spectra.from_spectrum(spectra['sfs_neut']),
            sfs_sel=Spectra.from_spectrum(spectra['sfs_sel'])
        )

    def create_polydfe_sfs_config(self, file: str):
        """
        Create a sfs config file for polyDFE.

        :param file: Path to the sfs config file to be created.
        """
        create_sfs_config(
            file=file,
            sfs_neut=self.data['sfs_neut'].all if isinstance(self.data['sfs_neut'], Spectra) else self.data['sfs_neut'],
            sfs_sel=self.data['sfs_sel'].all if isinstance(self.data['sfs_sel'], Spectra) else self.data['sfs_sel']
        )

    def to_dict(self) -> dict:
        """
        Represent config as dictionary.

        :return: Dictionary representation of config.
        """
        return self.data

    def to_json(self) -> str:
        """
        Create JSON representation of object.

        :return: JSON string
        """
        return json.dumps(self.data, indent=4, cls=CustomEncoder)

    def to_yaml(self) -> str:
        """
        Create YAML representation of object.

        :return: YAML string
        """
        return yaml.dump(json.loads(self.to_json()), sort_keys=False)

    def to_file(self, file: str):
        """
        Save object to file.

        :param file: Path to file.
        """
        with open(file, 'w') as fh:
            fh.write(self.to_yaml())

    @staticmethod
    def from_dict(data: dict) -> 'Config':
        """
        Load config from dictionary.

        :return: Config object.
        """
        # recreate spectra objects
        for sfs in ['sfs_neut', 'sfs_sel']:
            if sfs in data and data[sfs] is not None:
                data[sfs] = Spectra.from_dict(data=data[sfs])

        # recreate SharedParams objects
        if 'shared_params' in data and data['shared_params'] is not None:
            data['shared_params'] = [SharedParams(**p) for p in data['shared_params']]

        # recreate Covariate objects
        if 'covariates' in data and data['covariates'] is not None:
            data['covariates'] = [Covariate(**c) for c in data['covariates']]

        # cast to tuple of (float, float, int)
        # useful when restoring from YAML file
        for key in ['intervals_ben', 'intervals_del']:
            if key in data:
                data[key] = (float(data[key][0]), float(data[key][1]), int(data[key][2]))

        return Config(**data)

    @staticmethod
    def from_json(data: str) -> 'Config':
        """
        Load config from JSON str.

        :param data: JSON string.
        :return: Config object.
        """
        return Config.from_dict(json.loads(data))

    @staticmethod
    def from_yaml(data: str) -> 'Config':
        """
        Load config from YAML str.

        :param data: YAML string.
        :return: Config object.
        """
        return Config.from_dict(yaml.load(data, Loader=yaml.Loader))

    @classmethod
    def from_file(cls, file: str, cache: bool = True) -> 'Config':
        """
        Load object from file.

        :param file: Path to file, possibly gzipped.
        :param cache: Whether to use the cache if available.
        :return: Config object.
        """
        with open(download_if_url(file, cache=cache, desc=f'{cls.__name__}>Downloading file'), 'r') as fh:
            return Config.from_yaml(fh.read())

    def get_polydfe_model(self) -> str:  # pragma: no cover
        """
        Get the model name in polyDFE that corresponds
        to the configured DFE parametrization.

        :return: polyDFE model name.
        """
        # get name of configured model
        name = _to_string(self.data['model'])

        # return polyDFE model name if it exists
        if name in models:
            return models[name]

        # raise error otherwise
        raise NotImplementedError(f'There is no polyDFE equivalent of {name}')



"""
Visualization module.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import functools
import logging
from typing import Callable, List, Literal, Dict, Sequence

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import colors
from matplotlib.colors import LogNorm
from matplotlib.container import BarContainer
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .parametrization import Parametrization

# get logger
logger = logging.getLogger('fastdfe').getChild('Visualization')


class Visualization:
    """
    Visualization class.
    """

    @classmethod
    def change_default_figsize(cls, factor: float | np.ndarray):
        """
        Scale default figure size.

        :return: Factor to scale default figure size by
        """
        plt.rcParams["figure.figsize"] = list(factor * np.array(plt.rcParams["figure.figsize"]))

    @staticmethod
    def clear_show_save(func: Callable) -> Callable:
        """
        Decorator for clearing current figure in the beginning
        and showing or saving produced plot subsequently.

        :param func: Function to decorate
        :return: Wrapper function
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> plt.Axes:
            """
            Wrapper function.

            :param args: Positional arguments
            :param kwargs: Keyword arguments
            :return: Axes
            """

            # add axes if not given
            if 'ax' not in kwargs or ('ax' in kwargs and kwargs['ax'] is None):
                # clear current figure
                plt.close()

                kwargs['ax'] = plt.gca()

            # execute function
            func(*args, **kwargs)

            # make layout tight
            plt.tight_layout()

            # show or save
            # show by default here
            return Visualization.show_and_save(
                file=kwargs['file'] if 'file' in kwargs else None,
                show=kwargs['show'] if 'show' in kwargs else True
            )

        return wrapper

    @staticmethod
    def show_and_save(file: str = None, show: bool = True, pad: float = 1.08) -> plt.Axes:
        """
        Show and save plot.

        :param file: File path to save plot to
        :param show: Whether to show plot
        :param pad: Padding for tight layout
        :return: Axes

        """
        plt.tight_layout(pad=pad)

        # save figure if file path given
        if file is not None:
            plt.savefig(file, dpi=200, bbox_inches='tight', pad_inches=0.1)

        # show figure if specified and if not in interactive mode
        if show and not plt.isinteractive():
            plt.show()

        # return current axes
        return plt.gca()

    @staticmethod
    def interval_to_string(left: float, right: float) -> str:
        """
        Get string representation for given interval.

        :param left: Right interval boundary
        :param right: Left interval boundary
        :return: String representation of interval

        """
        # left interval is closed by default
        bracket_left = '[' if left != -np.inf else '('
        bracket_right = ')'

        def format_number(n: float) -> str:
            """
            Format number, allowing for np.inf.

            :param n: Number to format
            :return: Formatted number
            """
            if n == -np.inf:
                return '-inf'

            if n == np.inf:
                return 'inf'

            return '{:0.0f}'.format(n)

        return bracket_left + format_number(left) + ', ' + format_number(right) + bracket_right

    @staticmethod
    @clear_show_save
    def plot_discretized(
            ax: plt.Axes,
            values: Sequence,
            errors: Sequence = None,
            labels: Sequence = None,
            file: str = None,
            show: bool = True,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            title: str = 'discretized DFE',
            interval_labels: List[str] = None,
            kwargs_legend: dict = dict(prop=dict(size=8)),

    ) -> plt.Axes:
        """
        Plot discretized DFEs using a bar plot.

        :param interval_labels: Labels for the intervals (which are the same for all types)
        :param labels: Labels for the different types of DFEs
        :param title: Title of the plot
        :param values: Array of values of size ``intervals.shape[0] - 1``, containing the discretized DFE for each type
        :param errors: Array of errors of size ``intervals.shape[0] - 1``, containing the discretized DFE for each type
        :param file: File path to save plot to
        :param show: Whether to show plot
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bars.
        :param ax: Axes to plot on.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`.
        :return: Axes
        """
        # number of intervals
        n_intervals = len(intervals) - 1
        n_dfes = len(values)

        width_total = 0.9
        width = width_total / n_dfes

        for i in range(n_dfes):
            x = np.arange(n_intervals) + i * width

            # plot discretized DFE
            bars = ax.bar(
                x=x,
                height=values[i],
                width=width,
                yerr=errors[i] if errors is not None else None,
                error_kw=dict(
                    capsize=width * 7
                ),
                label=labels[i] if labels is not None else None,
                linewidth=0,
                hatch=Visualization.get_hatch(i, labels),
                color=Visualization.get_color(labels[i], labels) if labels is not None else None
            )

            Visualization.darken_edge_colors(bars)

        ax.set(xlabel='S', ylabel='fraction')

        # determine x-labels
        if interval_labels is None:
            xlabels = [Visualization.interval_to_string(intervals[i - 1], intervals[i]) for
                       i in range(1, n_intervals + 1)]
        else:
            xlabels = interval_labels

        # customize x-ticks
        x = np.arange(n_intervals)
        ax.set_xticks([i + (width_total - width) / 2 for i in x], x)
        ax.set_xticklabels(xlabels)

        # set title
        ax.set_title(title)

        # show legend if labels were given
        if labels is not None:
            ax.legend(**kwargs_legend)

        # remove x-margins
        ax.autoscale(tight=True, axis='x')

        return ax

    @staticmethod
    @clear_show_save
    def plot_continuous(
            ax: plt.Axes,
            bins: np.ndarray,
            values: Sequence,
            errors: Sequence = None,
            labels: Sequence = None,
            file: str = None,
            show: bool = True,
            title: str = 'continuous DFE',
            scale: Literal['log', 'linear'] = 'lin',
            ylim: float = 1e-2,
            scale_density: bool = False,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            **kwargs
    ) -> plt.Axes:
        """
        Plot DFEs using a line plot.
        By default, the PDF is plotted as is. Due to the logarithmic scale on
        the x-axis, we may get a wrong intuition on how the mass is distributed,
        however. To get a better intuition, we can optionally scale the density
        by the x-axis interval size using ``scale_density = True``. This has the
        disadvantage that the density now changes for x, so that even a constant
        density will look warped.

        :param ylim: y-axis limit
        :param scale: Scale of y-axis
        :param bins: Array of bin boundaries
        :param title: Title of the plot
        :param labels: Labels for the DFEs
        :param errors: Array of errors
        :param values: Array of values
        :param file: File path to save plot to
        :param show: Whether to show plot
        :param scale_density: Whether to scale the density by the bin size
        :param ax: Axes to plot on.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`.
        :return: Axes
        """
        from .discretization import get_midpoints_and_spacing

        n_bins = len(bins) - 1
        n_dfes = len(values)

        # get interval sizes
        _, interval_sizes = get_midpoints_and_spacing(bins)

        for i in range(n_dfes):
            x = np.arange(n_bins)

            # plot DFE
            ax.plot(x,
                    values[i] if scale_density else values[i] / interval_sizes,
                    label=labels[i] if labels is not None else None)

            # visualize errors
            if errors is not None and errors[i] is not None:
                ax.fill_between(
                    x=x,
                    y1=(values[i] - errors[i][0]) if scale_density else (values[i] - errors[i][0]) / interval_sizes,
                    y2=(values[i] + errors[i][1]) if scale_density else (values[i] + errors[i][1]) / interval_sizes,
                    alpha=0.2
                )

        # customize x-ticks
        Visualization.adjust_ticks_show_s(bins)

        # use log scale if specified
        if scale == 'log':
            ax.set_yscale('log')
            ax.set_ylim(bottom=ylim)

        # show legend if labels were given
        if labels is not None:
            ax.legend(**kwargs_legend)

        # remove x-margins
        ax.set_xmargin(0)

        ax.set_title(title)

        return ax

    @staticmethod
    def adjust_ticks_show_s(s: np.ndarray):
        """
        Adjust x-ticks to show bin values.

        :return: Array of selection coefficients
        """

        n_bins = len(s) - 1
        ax = plt.gca()

        ticks = ax.get_xticks()
        ticks_new = ["{:.0e}".format(s[int(l)]) if 0 <= int(l) < n_bins else None for l in ticks]

        ax.set_xticks(ticks)
        ax.set_xticklabels(ticks_new)

    @staticmethod
    def name_to_label(
            key: str,
            param_names: Sequence = None,
            vals: Sequence = None
    ) -> str:
        """
        Map parameter name to label.

        :param key: Parameter name
        :param param_names: Parameter names
        :param vals: Parameter values
        :return: Label
        """
        # define new names for some parameters
        label_mapping = dict(
            alpha='α',
            eps='ε'
        )

        # map to new name
        label = label_mapping[key] if key in label_mapping else key

        # wrap in math mode tags
        if param_names is None:
            return '$' + label + '$'

        # get index of parameter
        k = np.where(np.array(param_names) == key)[0][0]

        # check parameter value and add minus sign if negative
        if vals[k] >= 0:
            return '$' + label + '$'

        return '$-' + key + '$'

    @staticmethod
    @clear_show_save
    def plot_inferred_parameters(
            ax: plt.Axes,
            values: Dict[str, np.ndarray],
            labels: Sequence,
            param_names: Sequence,
            errors: Dict[str, np.ndarray | None],
            file: str = None,
            show: bool = True,
            title: str = 'parameter estimates',
            legend: bool = True,
            scale: Literal['lin', 'log', 'symlog'] = 'log',
            kwargs_legend: dict = dict(prop=dict(size=8), loc='upper right')
    ) -> plt.Axes:
        """
        Visualize the inferred parameters and their confidence intervals.
        using a bar plot. Note that there problems with parameters that span 0 (which is usually note the case).

        :param labels: Unique labels for the DFEs
        :param param_names: Labels for the parameters
        :param scale: Whether to use a linear or log scale
        :param title: Title of the plot
        :param legend: Whether to show the legend
        :param errors: Dictionary of errors with the parameter in the same order as ``labels``
        :param values: Dictionary of parameter values with the parameter in the same order as ``labels``
        :param file: File path to save plot to
        :param show: Whether to show plot
        :param ax: Axes to plot on.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`.
        :return: Axes
        """
        labels = list(labels)
        n_types = len(values)

        width_total = 0.9
        width = width_total / n_types

        # number of parameters
        n_params = len(param_names)

        # iterate over types
        for i, vals, errs in zip(range(n_types), values.values(), errors.values()):

            x = np.arange(n_params) + i * width

            # get labels for parameters
            xlabels = np.array([Visualization.name_to_label(key, param_names, vals) for key in param_names])

            # flip errors for negative parameters
            if errs is not None:
                errs = np.array([err if vals[i] >= 0 else err[::-1] for i, err in enumerate(errs.T)]).T

            # Plot bars.
            # Note that we plot the absolute value of the parameter
            bars = ax.bar(
                x=x,
                height=np.abs(vals),
                yerr=errs,
                error_kw=dict(
                    capsize=width * 7
                ),
                label=labels[i],
                width=width,
                linewidth=0,
                hatch=Visualization.get_hatch(i, labels),
                color=Visualization.get_color(labels[i], labels) if labels is not None else None
            )

            Visualization.darken_edge_colors(bars)

            # customize x-ticks
            x = np.arange(n_params)
            ax.set_xticks([i + (width_total - width) / 2 for i in x], x)
            ax.set_xticklabels(xlabels)

        # show legend if specified
        if legend:
            ax.legend(**kwargs_legend)

        # set title
        ax.set_title(title)

        # change to log-scale if specified
        if scale in ['log', 'symlog']:
            ax.set_yscale('symlog', linthresh=1e-3)

            # add y-margin at top
            ax.set_ylim(top=ax.get_ylim()[1] * 2)

        # remove x-margins
        ax.autoscale(tight=True, axis='x')

        return ax

    @staticmethod
    def plot_inferred_parameters_boxplot(
            values: Dict[str, pd.DataFrame],
            param_names: Sequence,
            file: str = None,
            show: bool = True,
            title: str = 'parameter estimates'
    ) -> plt.Axes:
        """
        Visualize the inferred parameters using a boxplot.

        :param values: Type-indexed dictionary of dataframes with parameters as columns and values as rows
        :param param_names: Parameters to plot
        :param title: Title of the plot
        :param file: File path to save plot to
        :param show: Whether to show plot
        :return: Axes
        """
        plt.close()

        # create a subplot with number of parameters axes
        fig, axs = plt.subplots(len(param_names), 1, figsize=(6, 1.4 * len(param_names)))

        axs[0].set_title(title)

        for i, param_name in enumerate(param_names):
            ax = axs[i]

            # Prepare a dataframe for seaborn
            df_list = []
            for t, df in values.items():
                subset = df[[param_name]].copy()
                subset['Type'] = t
                df_list.append(subset)

            df_plot = pd.concat(df_list)

            # plot
            sns.boxplot(x='Type', y=param_name, data=df_plot, ax=ax)

            ax.set_ylabel(
                Visualization.name_to_label(param_name, param_names, df_plot[param_name].abs().values),
                rotation=0,
                labelpad=20
            )

            ax.set_xlabel(None)

            if i < len(param_names) - 1:
                ax.set(xticklabels=[])
                ax.tick_params(bottom=False)
            else:
                [l.set_rotation(90) for l in ax.get_xticklabels()]

        # show and save plot
        return Visualization.show_and_save(file, show, pad=0.3)

    @staticmethod
    def get_color(
            label: str,
            labels: List[str],
            get_group: Callable = lambda x: x.split('.')[-1]
    ) -> str:
        """
        Get color for specified label.

        :param label: Label to get color for
        :param labels: List of labels
        :param get_group: Function to get group from label
        :return: Color string
        """
        # determine unique groups
        groups = np.unique(np.array([get_group(l) for l in labels]))

        # determine group of current label
        group = get_group(label)

        # determine index of group
        i = np.where(groups == group)[0][0] if sum(groups == group) > 0 else 0

        # return color
        return f'C{i}'

    @staticmethod
    def get_hatch(i: int, labels: List[str] = None) -> str | None:
        """
        Get hatch style for specified index i.

        :param labels: List of labels
        :param i: Index
        :return: Hatch style
        """

        # determine whether hatch style should be used
        if labels is None or len(labels) < 1 or '.' not in labels[i]:
            return

        # determine unique prefixes
        prefixes = set([label.split('.')[0] for label in labels if '.' in label])
        hatch_styles = ['/////', '\\\\\\\\\\', '***', 'ooo', 'xxx', '...']

        prefix = labels[i].split('.')[0]
        prefix_index = list(prefixes).index(prefix)

        return hatch_styles[prefix_index % len(hatch_styles)]

    @staticmethod
    def plot_spectra(
            ax: plt.Axes,
            spectra: List[List[float]] | np.ndarray,
            labels: List[str] | np.ndarray = [],
            colors: List[str] | np.ndarray = None,
            log_scale: bool = False,
            use_subplots: bool = False,
            show_monomorphic: bool = False,
            title: str = None,
            n_ticks: int = 10,
            file: str = None,
            show: bool = True,
            kwargs_legend: dict = dict(prop=dict(size=8))
    ) -> plt.Axes:
        """
        Plot the given 1D spectra.

        :param show_monomorphic: Whether to show monomorphic site counts
        :param n_ticks: Number of x-ticks to use
        :param ax: Axes to plot on.
        :param ax: Axes to plot on. Only for Python visualization backend and if ``use_subplots`` is ``False``.
        :param title: Title of plot
        :param spectra: List of lists of spectra or a 2D array in which each row is a spectrum in the
            same order as ``labels``
        :param colors: List of colors for each spectrum.
        :param labels: List of labels for each spectrum
        :param log_scale: Whether to use logarithmic y-scale
        :param use_subplots: Whether to use subplots
        :param file: File to save plot to
        :param show: Whether to show the plot
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`.
        :return: Axes
        """
        if len(spectra) == 0:
            logger.warning('No spectra to plot.')
            return ax

        if use_subplots:

            # clear current figure
            plt.close()

            n_plots = len(spectra)
            n_rows = int(np.ceil(np.sqrt(n_plots)))
            n_cols = int(np.ceil(np.sqrt(n_plots)))

            fig = plt.figure(figsize=(6.4 * n_cols ** (1 / 3), 4.8 * n_rows ** (1 / 3)))
            axes = fig.subplots(ncols=n_cols, nrows=n_rows, squeeze=False).flatten()

            # plot spectra individually
            for i in range(n_plots):
                Visualization.plot_spectra(
                    spectra=[spectra[i]],
                    labels=[labels[i]] if len(labels) else [],
                    colors=[colors[i]] if colors else None,
                    ax=axes[i],
                    n_ticks=15 // min(2, n_cols),
                    log_scale=log_scale,
                    show_monomorphic=show_monomorphic,
                    show=False
                )

                # set title
                axes[i].set_title(labels[i] if i < len(labels) else '')

            # make empty plots invisible
            [ax.set_visible(False) for ax in axes[n_plots:]]

            # show and save plot
            return Visualization.show_and_save(file, show)

        if ax is None:
            plt.close()
            _, ax = plt.subplots()

        # determine sample size and width
        n = len(spectra[0]) - 1
        width_total = 0.9
        width = width_total / len(spectra)

        x = np.arange(n + 1) if show_monomorphic else np.arange(1, n)

        # iterator over spectra and draw bars
        for i, sfs in enumerate(spectra):
            bars = ax.bar(
                x=x + i * width,
                height=sfs if show_monomorphic else sfs[1:-1],
                width=width,
                label=labels[i] if len(labels) else None,
                color=colors[i] if colors else None,
                linewidth=0,
                hatch=Visualization.get_hatch(i, labels)
            )

            Visualization.darken_edge_colors(bars)

        # adjust ticks
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        indices_ticks = x

        # filter ticks
        if n > n_ticks:
            indices_ticks = indices_ticks[indices_ticks % max(int(np.ceil(n / n_ticks)), 1) == 1]

        ax.set_xticks([i + (width_total - width) / 2 for i in indices_ticks], indices_ticks)

        ax.set_xlabel('frequency')

        # remove x-margins
        ax.autoscale(tight=True, axis='x')

        if log_scale:
            ax.set_yscale('log')

        # set title
        ax.set_title(title)

        # show legend if more than one label
        if len(labels) > 1:
            ax.legend(**kwargs_legend)

        # show and save plot
        return Visualization.show_and_save(file, show)

    @staticmethod
    def darken_edge_colors(bars: BarContainer):
        """
        Darken the edge color of the given bars.

        :param bars: Bars to darken
        """
        for bar in bars:
            color = bar.get_facecolor()
            edge_color = Visualization.darken_color(color, amount=0.75)
            bar.set_edgecolor(edge_color)

    @staticmethod
    def darken_color(color, amount=0.5) -> tuple:
        """
        Darken a color.

        :param color: Color to darken
        :param amount: Amount to darken
        :return: Darkened color as tuple
        """
        c = mcolors.to_rgba(color)

        return c[0] * amount, c[1] * amount, c[2] * amount, c[3]

    @staticmethod
    @clear_show_save
    def plot_pdf(
            model: Parametrization,
            params: dict,
            ax: plt.Axes,
            s: np.ndarray = np.linspace(-100, 100, 1000),
            file: str = None,
            show: bool = True
    ) -> plt.Axes:
        """
        Plot PDF of given parametrization.

        :param model: DFE parametrization
        :param params: Parameters to be used for parametrization
        :param s: Selection coefficients
        :param file: File to save plot to
        :param show: Whether to show plot
        :param ax: Axes to plot on.
        :return: Axes
        """
        ax.plot(np.arange(len(s)), model.get_pdf(**params)(s))

        # customize x-ticks
        Visualization.adjust_ticks_show_s(s)

        # remove x-margins
        ax.margins(x=0)

        return ax

    @staticmethod
    @clear_show_save
    def plot_cdf(
            model: Parametrization,
            params: dict,
            ax: plt.Axes,
            s: np.ndarray = np.linspace(-100, 100, 1000),
            file: str = None,
            show: bool = True
    ) -> plt.Axes:
        """
        Plot CDF of given parametrization.

        :param model: DFE parametrization
        :param params: Parameters to be used for parametrization
        :param s: Selection coefficients
        :param ax: Axes to plot on.
        :param file: File to save plot to
        :param show: Whether to show plot
        :return: Axes
        """
        ax.plot(np.arange(len(s)), model.get_cdf(**params)(s))

        # customize x-ticks
        Visualization.adjust_ticks_show_s(s)

        # remove x-margins
        ax.margins(x=0)

        return ax

    @staticmethod
    @clear_show_save
    def plot_scatter(
            values: Sequence,
            file: str,
            show: bool,
            ax: plt.Axes,
            title: str | None = None,
            scale: Literal['lin', 'log', 'symlog'] = 'lin',
            ylabel: str = 'lnl'
    ) -> plt.Axes:
        """
        A scatter plot.

        :param scale: Scale of y-axis
        :param values: Values to plot
        :param file: File to save plot to
        :param show: Whether to show plot
        :param title: Title of plot
        :param ax: Axes to plot on.
        :param ylabel: Label of y-axis
        :return: Axes
        """
        # plot
        sns.scatterplot(x=range(len(values)), y=values, ax=ax)

        ax.set(ylabel=ylabel)

        # set title
        ax.set_title(title)

        if scale == 'log':
            ax.set_yscale('symlog')

        return ax

    @staticmethod
    @clear_show_save
    def plot_buckets_sizes(
            n_intervals: int,
            bins: Sequence,
            sizes: Sequence,
            title: str,
            file: str,
            show: bool,
            ax: plt.Axes
    ) -> plt.Axes:
        """
        A line plot of the bucket sizes.

        :param bins: Bins of the histogram
        :param n_intervals: Number of intervals
        :param sizes: Sizes of the buckets
        :param title: Title of plot
        :param file: File to save plot to
        :param show: Whether to show plot
        :param ax: Axes to plot on.
        :return: Axes
        """
        # plot line
        ax.plot(np.arange(n_intervals), sizes)

        # use log scale
        ax.set_yscale('log')
        ax.set_title(title)

        # customize x-ticks
        Visualization.adjust_ticks_show_s(bins)

        # remove x-margins
        ax.set_xmargin(0)

        return ax

    @staticmethod
    @clear_show_save
    def plot_nested_models(
            P: np.ndarray,
            labels_x: Sequence,
            labels_y: Sequence,
            ax: plt.Axes,
            file: str = None,
            show: bool = True,
            cmap: str = None,
            title: str = None,
            vmin: float = 1e-10,
            vmax: float = 1,
    ) -> plt.Axes:
        """
        Plot p-values of nested likelihoods.

        :param P: Matrix of p-values
        :param labels_x: Labels for x-axis
        :param labels_y: Labels for y-axis
        :param file: File to save plot to
        :param show: Whether to show plot
        :param cmap: Colormap to use
        :param title: Title of plot
        :param ax: Axes to plot on.
        :param vmin: Minimum value for colorbar
        :param vmax: Maximum value for colorbar
        :return: Axes
        """

        def format_number(x: float | int | None) -> float | int | str:
            """
            Format number to be displayed.

            :param x: Number to format
            :return: Formatted number
            """
            if x == 0 or x is None:
                return 0

            if x < 0.0001:
                return "{:.1e}".format(x)

            return np.round(x, 4)

        # determine values to display
        annotation = np.vectorize(lambda x: str(format_number(x)))(P)
        annotation[np.equal(P, None)] = '-'

        # change to 1 to get a nicer color
        P[np.equal(P, None)] = 1

        # keep within color bar bounds
        P[P < vmin] = vmin
        P[P > vmax] = vmax

        # make the cbar have the same height as the heatmap
        cbar_ax = make_axes_locatable(ax).new_horizontal(size="4%", pad=0.15)
        plt.gcf().add_axes(cbar_ax)

        # default color map
        if cmap is None:
            cmap = colors.LinearSegmentedColormap.from_list('_', plt.get_cmap('inferno')(np.linspace(0.3, 1, 100)))

        # plot heatmap
        sns.heatmap(
            P.astype(float),
            ax=ax,
            cbar_ax=cbar_ax,
            cmap=cmap,
            norm=LogNorm(
                vmin=vmin,
                vmax=vmax
            ),
            annot=annotation,
            fmt="",
            square=True,
            linewidths=0.5,
            linecolor='#cccccc',
            cbar_kws=dict(
                label='p-value'
            )
        )

        # adjust tick labels
        ax.set_xticklabels([l.replace('_', ' ') for l in labels_x], rotation=45)
        ax.set_yticklabels([l.replace('_', ' ') for l in labels_y], rotation=0)

        # set title
        ax.set_title(title)

        return ax

    @staticmethod
    @clear_show_save
    def plot_interval_density(
            ax: plt.Axes,
            density: np.ndarray,
            intervals: Sequence = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            interval_labels: List[str] = None,
            file: str = None,
            show: bool = True,
            color: str = 'C0',
    ) -> plt.Axes:
        """
        Plot density of the discretization intervals chosen.

        :param density: Discretized density.
        :param color: Color of the bars.
        :param file: File to save plot to.
        :param show: Whether to show plot.
        :param interval_labels: Labels for the intervals.
        :param intervals: Array of interval boundaries yielding ``intervals.shape[0] - 1`` bins.
        :param ax: Axes to plot on.
        :return: Axes object
        """
        # number of intervals
        n_intervals = len(intervals)

        # x-values
        x = np.arange(n_intervals - 1)

        # plot plot
        sns.barplot(x=x, y=density / density.sum(), color=color, ax=ax)

        # set labels
        ax.set(xlabel='S', ylabel='fraction')

        # determine x-labels
        if interval_labels is None:
            xlabels = [Visualization.interval_to_string(intervals[i - 1], intervals[i]) for i in range(1, n_intervals)]
        else:
            xlabels = interval_labels

        # adjust x-ticks
        ax.set_xticks(x)
        ax.set_xticklabels(xlabels)

        # set title
        ax.set_title('interval density')

        # remove x-margins
        ax.autoscale(tight=True, axis='x')

        return ax

    @staticmethod
    @clear_show_save
    def plot_covariate(
            covariates: Sequence,
            values: Sequence,
            xlabel: str,
            ylabel: str,
            labels: Sequence = None,
            errors: Sequence = None,
            file: str = None,
            show: bool = True,
            title: str = 'likelihoods',
            ax: plt.Axes = None
    ) -> plt.Axes:
        """

        :param covariates: The covariate values.
        :param values: The MLE parameter values.
        :param xlabel: X-axis label.
        :param ylabel: Y-axis label.
        :param labels: Labels for the different types
        :param errors: The errors on the MLE parameter values.
        :param labels: Labels for the different types
        :param file: File to save plot to
        :param show: Whether to show plot
        :param title: Title of plot
        :param ax: Axes to plot on.
        :return: Axes
        """
        ax.scatter(x=covariates, y=np.array(values))

        # plot error bars
        if errors is not None:
            ax.errorbar(
                x=covariates,
                y=np.array(values),
                yerr=np.array(errors),
                fmt='none',
                capsize=3
            )

        # set labels
        ax.set(xlabel=xlabel, ylabel=Visualization.name_to_label(ylabel))

        # set title
        ax.set_title(title)

        # Add second x-axis on top of the plot to show the labels
        if labels is not None:
            ax2 = ax.twiny()
            ax2.set_xticks(covariates)
            ax2.set_xticklabels(labels, rotation=45, ha='left')
            ax2.set_xlim(ax.get_xlim())

        return ax
